import json
import logging
import re
from collections import OrderedDict
from importlib import resources
from typing import List, Optional

import pandas as pd
from cas_schema import schemas

from cas.anndata_to_cas import calculate_labelset
from cas.cxg_utils import download_dataset_with_id
from cas.file_utils import read_anndata_file
from cas.utils.conversion_utils import (
    add_labelsets_to_cas,
    add_parent_cell_hierarchy,
    add_parent_hierarchy_to_annotations,
    generate_parent_cell_lookup,
    get_cell_ids,
)
from cas.file_utils import get_cas_schema_names

logging.basicConfig(level=logging.INFO)

LABELSET_COLUMN = "CELL LABELSET NAME"
CELL_LABEL_COLUMN = "CELL TYPE TERM"
CL_TERM_COLUMN = "CL TERM"
EVIDENCE_COLUMN = "EVIDENCE"
MARKER_GENES_COLUMN = "MARKER GENES"
SYNONYM_COLUMN = "SYNONYMS"
CATEGORIES_COLUMN = "CATEGORIES"


def resolve_ref(schema, ref):
    parts = ref[1:].split("/")
    definition = schema
    for part in parts[1:]:
        definition = definition[part]
    return definition


def read_spreadsheet(file_path: str, sheet_name: Optional[str], schema: dict):
    """
    Read the specific sheet from the Excel file into a pandas DataFrame.

    Args:
        file_path (str): Path to the Excel file.
        sheet_name (str, optional): Target sheet name. If not provided, reads the first sheet.
        schema: Cell annotation schema

    Returns:
        tuple: Tuple containing metadata (dict), column names (list), and raw data (pd.DataFrame).
    """
    if sheet_name:
        spreadsheet_df = pd.read_excel(file_path, sheet_name=sheet_name, header=None)
    else:
        spreadsheet_df = pd.read_excel(file_path, header=None)

    meta_data = {}

    header_row_index = None
    metadata_properties = {
        k: v
        for k, v in schema["properties"].items()
        if k not in ["labelsets", "annotations"]
    }
    for index, row in spreadsheet_df.iterrows():
        first_cell = str(row[0])
        if first_cell.startswith("#"):
            key = first_cell[1:].strip()
            value = row[1] if pd.notnull(row[1]) else ""
            if key in metadata_properties:
                meta_data[key] = value
                if metadata_properties[key]["type"] == "array":
                    meta_data[key] = re.split("[,|]", value)
        else:
            header_row_index = index
            break

    if header_row_index is not None:
        column_names = spreadsheet_df.iloc[header_row_index, :].tolist()
        raw_data = spreadsheet_df.iloc[header_row_index + 1 :, :]
        raw_data.columns = column_names
        # Iterate through each column and filter out rows with 'cell_type' in any column
        for column in raw_data.columns:
            raw_data = raw_data[raw_data[column] != "cell_type"]
        raw_data = raw_data.where(pd.notnull(raw_data), "")
    else:
        raise ValueError("Header row not found in the spreadsheet.")

    return meta_data, column_names, raw_data


def retrieve_schema(schema_name):
    schema_name = str(schema_name).strip().lower()
    if schema_name not in get_cas_schema_names():
        raise Exception("Schema name should be one of 'base', 'bican' or 'cap'")
    schema_file = resources.files(schemas) / get_cas_schema_names()[schema_name]
    with schema_file.open("rt") as f:
        schema = json.loads(f.read())
    return schema


def custom_lowercase_transform(s):
    """
    Transforms the given string to lowercase except for words that are acronyms or specific cell type names
    which are three characters or fewer.

    Args:
        s (str): The input string.

    Returns:
        str: The transformed string.
    """

    # Define a function to decide whether a word should stay uppercase
    def transform_word(match):
        word = match.group()
        # If a word is three characters or fewer, it stays uppercase
        if len(word) <= 3:
            return word
        # Otherwise, convert the word to lowercase
        else:
            return word.lower()

    # Use regex to find words and apply the transformation logic
    transformed_string = re.sub(r"\b[A-Za-z]+\b", transform_word, s)

    return transformed_string


def spreadsheet2cas(
    spreadsheet_file_path: str,
    sheet_name: Optional[str],
    anndata_file_path: Optional[str],
    labelset_list: Optional[List[str]],
    schema_name: Optional[str],
    output_file_path: str,
):
    """
    Convert a spreadsheet to Cell Annotation Schema (CAS) JSON.

    Args:
        spreadsheet_file_path (str): Path to the spreadsheet file.
        sheet_name (Optional[str]): Target sheet name in the spreadsheet. Can be a string or None.
        anndata_file_path: The path to the AnnData file.
        labelset_list (Optional[List[str]]): List of names of observation (obs) fields used to record author cell
        type names, which determine the rank of labelsets in a spreadsheet.
        schema_name (Optional[str]): Name of the CAS schema, can be one of 'base', 'bican' or 'cap'.
        output_file_path (str): Output CAS file name.
    """
    cell_annotation_schema = retrieve_schema(schema_name if schema_name else "cap")
    meta_data_result, column_names_result, raw_data_result = read_spreadsheet(
        spreadsheet_file_path, sheet_name, cell_annotation_schema
    )

    anndata, matrix_file_id = load_or_fetch_anndata(anndata_file_path, meta_data_result)

    if not labelset_list:
        labelset_list = raw_data_result["labelset"].unique().tolist()

    labelset_dict = calculate_labelset(anndata.obs, labelset_list)

    cas = initialize_cas_structure(matrix_file_id, meta_data_result)

    add_labelsets_to_cas(cas, labelset_dict)

    parent_cell_look_up = generate_parent_cell_lookup(anndata, labelset_dict)

    add_annotations_to_cas(
        cas,
        raw_data_result,
        column_names_result,
        cell_annotation_schema,
        parent_cell_look_up,
    )

    add_parent_cell_hierarchy(parent_cell_look_up)
    add_parent_hierarchy_to_annotations(cas, parent_cell_look_up)

    # Write the JSON data to the file
    with open(output_file_path, "w") as json_file:
        json.dump(cas, json_file, indent=2)


def add_annotations_to_cas(cas, raw_data_result, columns, schema, parent_cell_look_up):
    """
    Adds processed annotations from raw data to the CAS structure and tracks labelsets.
    Assumes certain external definitions for column names and transformation functions.

    Args:
        cas (dict): The CAS structure to update with annotations.
        raw_data_result (DataFrame): Raw annotation data.
        columns (list): Column names of raw data to process.
        schema (dict): Cell annotation schema.
        parent_cell_look_up (Dict[str, Any]): A precomputed dictionary containing hierarchical metadata about cell
            labels.

    Returns:
        OrderedDict: Tracks labelsets encountered, initialized to None.

    Note:
        Requires `custom_lowercase_transform`, `get_cell_ids`, and column constants to be defined.
    """
    stripped_data_result = raw_data_result.map(
        lambda x: x.strip() if isinstance(x, str) else x
    )
    for index, row in stripped_data_result.iterrows():
        anno = {}
        user_annotations = {}
        label = row["cell_label"]
        annotation_properties = schema["definitions"]["Annotation"]["properties"]

        for column_name in columns:
            if column_name in annotation_properties:
                anno[column_name] = row[column_name]
                if column_name == "labelset":
                    labelset = anno[column_name]
                if annotation_properties[column_name]["type"] == "array":
                    anno[column_name] = re.split("[,|]", row[column_name])
            elif row[column_name]:
                user_annotations[column_name] = row[column_name]

        anno.update(
            {
                "cell_ids": list(parent_cell_look_up[f"{labelset}:{label}"]["cell_ids"]),
                "cell_set_accession": parent_cell_look_up[f"{labelset}:{label}"]["accession"],
                "cell_ontology_term_id": parent_cell_look_up[f"{labelset}:{label}"][
                    "cell_ontology_term_id"
                ],
                "cell_ontology_term": parent_cell_look_up[f"{labelset}:{label}"]["cell_ontology_term"],
            }
        )
        if user_annotations:
            anno["author_annotation_fields"] = user_annotations
        cas.get("annotations").append(anno)


def initialize_cas_structure(matrix_file_id: str, meta_data_result: dict):
    """
    Initializes the Cell Annotation Schema (CAS) structure with basic information and placeholders
    for annotations and labelsets. Fields initialized with None values are omitted in the final output.

    Args:
        matrix_file_id (str): The ID of the matrix file, used within the CAS for identification.
        meta_data_result (dict): Metadata containing at least the 'matrix_file_id' for the CAS URL.

    Returns:
        dict: The initial CAS structure with the matrix file ID, annotation URL, and placeholders
              for future data. Excludes fields that remain None.
    """
    cas_init = {k: v for k, v in meta_data_result.items()}
    cas_init.update(
        {
            "matrix_file_id": f"cxg_dataset:{matrix_file_id}",
            "cellannotation_url": meta_data_result["matrix_file_id"],
            "annotations": [],
            "labelsets": [],
        }
    )
    cas = {k: v for k, v in cas_init.items() if v is not None}
    return cas


def load_or_fetch_anndata(anndata_file_path: str, meta_data_result: dict):
    """
    Loads or fetches an AnnData file, based on a local path or a matrix file ID from metadata.

    Args:
        anndata_file_path (str): Path to an AnnData file, or None to fetch using metadata.
        meta_data_result (dict): Metadata with 'matrix_file_id' for fetching the dataset.

    Returns:
        tuple: (AnnData object, matrix file ID), ready for use.

    Raises:
        ValueError: If 'matrix_file_id' is missing from metadata.
    """
    matrix_file_id = (
        meta_data_result["matrix_file_id"].rstrip("/").split("/")[-1].split(".")[0]
    )
    if not anndata_file_path:
        anndata_file_path = download_dataset_with_id(matrix_file_id)
    dataset_anndata = read_anndata_file(anndata_file_path)
    return dataset_anndata, matrix_file_id



import json
from typing import Optional

import anndata

from cas.file_utils import read_anndata_file, read_json_file


def populate_cell_ids(cas_json_path: str, anndata_path: str, labelsets: list = None):
    """
    Add/update CellIDs to CAS from matching AnnData file.

    Parameters:
        cas_json_path: path to CAS JSON file
        anndata_path: path to anndata file
        labelsets: List of labelsets to update with IDs from AnnData. If value is null, rank '0' labelset is used.
    """
    ad = read_anndata_file(anndata_path)
    if ad is not None:
        cas = read_json_file(cas_json_path)
        cas = add_cell_ids(cas, ad, labelsets)
        if cas:
            with open(cas_json_path, "w") as json_file:
                json.dump(cas, json_file, indent=2)
    else:
        raise Exception("Anndata read operation failed: {}".format(anndata_path))


def add_cell_ids(cas: dict, ad: Optional[anndata.AnnData], labelsets: list = None):
    """
    Add/update CellIDs to CAS from matching AnnData file.

    Parameters:
        cas: CAS JSON object
        ad: anndata object
        labelsets: List of labelsets to update with IDs from AnnData. If value is null, rank '0' labelset is used. The
        labelsets should be provided in order, starting from rank 0 (leaf nodes) and ascending to higher ranks.
    """
    rank_zero_labelset = [lbl_set["name"] for lbl_set in cas["labelsets"]
                          if isinstance(cas["labelsets"][0].get("rank"), int) and lbl_set.get("rank") == 0
                          or lbl_set.get("rank") == "0"][0]
    if not labelsets:
        labelsets = rank_zero_labelset

    cluster_identifier_column = get_obs_cluster_identifier_column(ad, labelsets, rank_zero_labelset)

    if cluster_identifier_column:
        cid_lookup = {}
        for anno in cas["annotations"]:
            if anno["labelset"] == rank_zero_labelset and anno["labelset"] in labelsets:
                cell_ids = []
                if cluster_identifier_column.endswith("_id") or cluster_identifier_column.lower().endswith(" id"):
                    # cluster column value is integer cluster id
                    cluster_id = anno.get("author_annotation_fields", {}).get(cluster_identifier_column)
                    if not cluster_id:
                        raise ValueError("AnnData cluster identifier column ({}) couldn't be find found in"
                                         "CAS author_annotation_fields.".format(cluster_identifier_column))
                    cell_ids = list(
                        ad.obs.loc[
                            ad.obs[cluster_identifier_column] == int(cluster_id),
                            cluster_identifier_column,
                        ].index
                    )
                else:
                    # cluster column value is cell label
                    cluster_label = anno["cell_label"]
                    cell_ids = list(
                        ad.obs.loc[
                            ad.obs[cluster_identifier_column] == cluster_label
                            ].index
                    )
                anno["cell_ids"] = cell_ids
                if "parent_cell_set_name" in anno:
                    lookup_key = anno["parent_cell_set_name"]
                    if lookup_key in cid_lookup:
                        cid_lookup[lookup_key].update(cell_ids)
                    else:
                        cid_lookup[lookup_key] = set(cell_ids)

        for anno in cas["annotations"]:
            if anno["labelset"] in labelsets and anno["cell_label"] in cid_lookup:
                cell_ids = list(cid_lookup.get(anno["cell_label"], []))
                anno["cell_ids"] = cell_ids

        return cas
    else:
        print(
            "WARN: Cluster identifier column couldn't be identified in OBS. Populate cell ids operation is aborted."
        )

    return None


def get_obs_cluster_identifier_column(ad, labelsets: list = None, rank_zero_labelset: str = None):
    """
    Anndata files may use different column names to uniquely identify Clusters. Get the cluster identifier column name for the current file.
    Args:
        ad: anndata object
        labelsets: List of labelsets to update with IDs from AnnData. The labelsets should be provided in order,
        starting from rank 0 (leaf nodes) and ascending to higher ranks.
        rank_zero_labelset: rank 0 labelset name
    Returns:
        cluster identifier column name
    """
    obs_keys = ad.obs_keys()
    cluster_identifier_column = ""
    if labelsets and labelsets[0] in obs_keys:
        cluster_identifier_column = labelsets[0]
    else:
        # check id variants
        if rank_zero_labelset.title() + "_id" in obs_keys:
            cluster_identifier_column = rank_zero_labelset.title() + "_id"
        elif rank_zero_labelset.lower() + "_id" in obs_keys:
            cluster_identifier_column = rank_zero_labelset.lower() + "_id"
        elif rank_zero_labelset.title() + " id" in obs_keys:
            cluster_identifier_column = rank_zero_labelset.title() + " id"
        elif rank_zero_labelset.lower() + " id" in obs_keys:
            cluster_identifier_column = rank_zero_labelset.lower() + " id"
    return cluster_identifier_column


# def populate_cell_ids(cas_json_path: str, anndata_path: str, labelsets: list = None):
#     """
#     Add/update CellIDs to CAS from matching AnnData file.
#
#     Parameters:
#         cas_json_path: path to CAS JSON file
#         anndata_path: path to anndata file
#         labelsets: List of labelsets to update with IDs from AnnData. If value is null, rank '0' labelset is used.
#     """
#     ad = read_anndata_file(anndata_path)
#     cas = read_json_file(cas_json_path)
#
#     if not labelsets:
#         labelsets = [lbl_set["name"] for lbl_set in cas["labelsets"] if lbl_set["rank"] == "0"]
#
#     print(labelsets)
#     out = {}
#     for k, v in ad.obs.iterrows():
#         for r in labelsets:
#             val = v[r]
#             if val:
#                 if not r in out.keys(): out[r] = {}
#                 if not val in out[r].keys(): out[r][val] = []
#                 out[r][v[r]].append(k)
#
#     for a in cas['annotations']:
#         if not (a['labelset'] in out.keys()):
#             print("Unknown labelset %s", (a['labelset']))
#         elif not (a['cell_label']) in out[a['labelset']].keys():
#             print("Unknown value %s in labelset %s", (a['cell_label'], a['labelset']))
#         else:
#             a['cell_ids'] = out[a['labelset']][a['cell_label']]
#
#     with open(cas_json_path, 'w') as f:
#         f.write(json.dumps(cas))
#
#     ad.file.close()



import json
import os
from importlib.metadata import version
from typing import Any, Dict

import pandas as pd

from cas.file_utils import read_json_file

CAT_SET_REQUIRED_COLUMNS = ["label", "name","description", "order"]
CAT_REQUIRED_COLUMNS = [
    "cluster_annotation_term_set_label",
    "name",
    "label",
    "parent_term_label",
    "cluster_annotation_term_set_name",
]


def validate_dataframe_columns(df: pd.DataFrame, required_columns: list):
    """
    Validates a DataFrame for the required columns.
    Args:
        df (pandas.DataFrame): DataFrame to validate.
        required_columns (list): List of required column names.
    """
    missing_columns = [col for col in required_columns if col not in df.columns]
    if missing_columns:
        raise ValueError(f"Missing columns in DataFrame: {', '.join(missing_columns)}")


def generate_catset_dataframe(cas: Dict[str, Any]) -> pd.DataFrame:
    """
    Generate a DataFrame representing the Cluster Annotation Term Set (cat_set)
    from the given Cell Annotation Schema (CAS) dictionary.

    Args:
        cas (Dict[str, Any]): The Cell Annotation Schema (CAS) dictionary.

    Returns:
        pd.DataFrame: DataFrame representing the Cluster Annotation Term Set (cat_set).

    """
    labelsets = cas.get("labelsets", [])
    data = [
        {
            "name": labelset.get("name"),
            "description": labelset.get("description"),
            "order": labelset.get("rank"),
        }
        for labelset in labelsets
    ]
    catset_df = pd.DataFrame(data)
    return catset_df


def generate_cat_dataframe(cas: Dict[str, Any]) -> pd.DataFrame:
    """
    Generate a DataFrame representing the Cluster Annotation Term (cat) from the given Cell Annotation Schema (CAS)
    dictionary.

    Args:
        cas (Dict[str, Any]): The Cell Annotation Schema (CAS) dictionary.

    Returns:
        pd.DataFrame: DataFrame representing the Cluster Annotation Term (cat).
    """
    columns = [
        "label",
        "name",
        "cluster_annotation_term_set_label",
        "parent_term_label",
        "parent_term_set_label",
        "term_set_order",
        "term_order",
        "cluster_annotation_term_set_name",
    ]

    annotations = cas.get("annotations")
    data = []
    for annotation in annotations:
        cell_label = annotation.get("cell_label")
        cell_set_accession = annotation.get("cell_set_accession")
        parent_cell_set_accession = annotation.get("parent_cell_set_accession")
        new_row = {
            "label": cell_set_accession,
            "name": cell_label,
            "parent_term_label": parent_cell_set_accession,
        }
        data.append(new_row)

    cat_df = pd.DataFrame(data, columns=columns)
    return cat_df


def calculate_order_mapping(order_values: pd.Series) -> Dict[str, str]:
    """
    Calculate a mapping dictionary based on the order values.

    Args:
        order_values (pandas.Series): Series containing the order values.

    Returns:
        Dict[str, str]: Mapping dictionary where keys are order values and values are rank values.
    """
    order_values_filtered = order_values[order_values != 0]
    order_values_sorted = order_values_filtered.sort_values(ascending=False)
    mapping = dict(zip(order_values_sorted, range(len(order_values_sorted))))

    return mapping


def abc2cas(cat_set_file_path: str, cat_file_path: str, output_file_path: str):
    """
    Converts given ABC files to a Cell Annotation Schema (CAS) JSON and writes it to a file with output_file_path name.
    Args:
        cat_set_file_path: Path to the Cluster Annotation Term Set file.
        cat_file_path: Path to the Cluster Annotation Term file.
        output_file_path: Output CAS file name (default: output.json).

    """
    cat_set = pd.read_csv(cat_set_file_path, sep=",")
    cat = pd.read_csv(cat_file_path, sep=",")

    validate_dataframe_columns(cat_set, CAT_SET_REQUIRED_COLUMNS)
    validate_dataframe_columns(cat, CAT_REQUIRED_COLUMNS)

    cas = init_metadata()
    add_labelsets(cas, cat_set)
    add_annotations(cas, cat)

    # Write the JSON data to the file
    with open(output_file_path, "w") as json_file:
        json.dump(cas, json_file, indent=2)


def add_annotations(cas: Dict[str, Any], cat: pd.DataFrame):
    """
    Adds annotations to the Cell Annotation Schema (CAS) based on the data from the Cluster Annotation Term DataFrame.

    Args:
        cas (Dict[str, Any]): Dictionary representing the Cell Annotation Schema.
        cat (pd.DataFrame): DataFrame containing Cluster Annotation Term data.

    """
    for row in cat.itertuples():
        labelset = row.cluster_annotation_term_set_name
        cell_label = row.name
        cell_fullname = None
        cell_ontology_term_id = None
        cell_ontology_term = None
        cell_ids = None
        rationale = None
        rationale_dois = None
        marker_gene_evidence = None
        synonyms = None
        category_fullname = None
        category_cell_ontology_exists = None
        category_cell_ontology_term_id = None
        category_cell_ontology_term = None
        cell_set_accession = row.label
        parent_cell_set_accession = row.parent_term_label

        anno = {
            "labelset": labelset,
            "cell_label": cell_label,
            "cell_fullname": cell_fullname,
            "cell_ontology_term_id": cell_ontology_term_id,
            "cell_ontology_term": cell_ontology_term,
            "cell_ids": cell_ids,
            "rationale": rationale,
            "rationale_dois": rationale_dois,
            "marker_gene_evidence": marker_gene_evidence,
            "synonyms": synonyms,
            "category_fullname": category_fullname,
            "category_cell_ontology_exists": category_cell_ontology_exists,
            "category_cell_ontology_term_id": category_cell_ontology_term_id,
            "category_cell_ontology_term": category_cell_ontology_term,
            "cell_set_accession": cell_set_accession,
            "parent_cell_set_accession": parent_cell_set_accession,
        }
        cas.get("annotations").append({k: v for k, v in anno.items() if v is not None and not pd.isna(v)})


def add_labelsets(cas: Dict[str, Any], cat_set: pd.DataFrame):
    """
    Adds labelsets to the Cell Annotation Schema (CAS) based on the data from the Cluster Annotation Term Set DataFrame.

    Args:
        cas (Dict[str, Any]): Cell Annotation Schema dictionary.
        cat_set (pandas.DataFrame): DataFrame containing Cluster Annotation Term Set data.
    """
    order_mapping = calculate_order_mapping(cat_set["order"])
    for row in cat_set.itertuples():
        name = row.name
        description = row.description
        rank = None
        if row.order != 0:
            rank = order_mapping[row.order]

        labelset = {"name": name, "description": description}
        if rank is not None:
            labelset["rank"] = rank
        cas.get("labelsets").append(labelset)


def init_metadata() -> Dict[str, Any]:
    """
    Initializes metadata for Cell Annotation Schema (CAS).

    Returns:
        Dict[str, Any]: Metadata dictionary containing default values for various fields.
    """
    # TODO These needs proper assignments
    matrix_file_id = None
    cellannotation_schema_version = version("cell-annotation-schema")
    cellannotation_timestamp = None
    cellannotation_version = None
    cellannotation_url = None
    author_name = "Jane Doe"  # TODO Needs proper author_name assignment
    author_contact = None
    orcid = None
    cas = {
        "matrix_file_id": matrix_file_id,
        "cellannotation_schema_version": cellannotation_schema_version,
        "cellannotation_timestamp": cellannotation_timestamp,
        "cellannotation_version": cellannotation_version,
        "cellannotation_url": cellannotation_url,
        "author_name": author_name,
        "author_contact": author_contact,
        "orcid": orcid,
        "annotations": [],
        "labelsets": [],
    }
    cas = {k: v for k, v in cas.items() if v is not None}
    return cas


def cas2abc(cas_file_path: str, cat_set_file_path: str, cat_file_path: str):
    """
    Converts given Cell Annotation Schema (CAS) to ABC files: cluster_annotation_term and
    cluster_annotation_term_set, and writes them to files with cat_file_path and cat_set_file_path.

    Args:
        cas_file_path: Path to the Cell Annotation Schema (CAS) file
        cat_set_file_path: Path to the Cluster Annotation Term Set file.
        cat_file_path: Path to the Cluster Annotation Term file.

    """
    cas_json = read_json_file(cas_file_path)

    cluster_annotation_term_set = generate_catset_dataframe(cas_json)
    cluster_annotation_term = generate_cat_dataframe(cas_json)

    current_directory = os.getcwd()
    cluster_annotation_term_set.to_csv(
        os.path.join(current_directory, cat_set_file_path), index=False
    )
    cluster_annotation_term.to_csv(
        os.path.join(current_directory, cat_file_path), index=False
    )
    # TODO implement rest of the method once the requirements are more clear



"""
cxg_utils.py

This module provides utility functions for working with AnnData datasets in the context of the CellxGene Census library.
"""

import logging
import os
from typing import Optional

import cellxgene_census


def download_dataset_with_id(dataset_id: str, file_path: Optional[str] = None) -> str:
    """
    Download an AnnData dataset with the specified ID.

    Args:
        dataset_id (str): The ID of the dataset to download.
        file_path (Optional[str], optional): The file path to save the downloaded AnnData. If not provided,
            the dataset will be saved in the current working directory with the dataset_id as the file name.
            Supports both absolute and relative paths.

    Returns:
        str: The path to the downloaded AnnData dataset
    """
    default_file_name = f"{dataset_id}.h5ad"
    anndata_file_path = default_file_name if file_path is None else file_path

    anndata_file_path = os.path.abspath(anndata_file_path)

    # Check if the file already exists
    if os.path.exists(anndata_file_path):
        print(f"File '{anndata_file_path}' already exists. Skipping download.")
        return anndata_file_path

    # Ensure the directory exists
    directory = os.path.dirname(anndata_file_path)
    if directory and not os.path.exists(directory):
        os.makedirs(directory, exist_ok=True)

    logging.info(f"Downloading dataset with ID '{dataset_id}'...")
    cellxgene_census.download_source_h5ad(dataset_id, to_path=anndata_file_path)
    logging.info(f"Download complete. File saved at '{anndata_file_path}'.")
    return anndata_file_path



import itertools
from datetime import date, datetime
import requests
from typing import Any, Dict, List, Tuple

import anndata as ad
import pandas as pd

from cas.accession.hash_accession_manager import HashAccessionManager


CROSSREF_API_URL = "https://api.crossref.org/works/"


def calculate_labelset_rank(input_list: List[str]) -> Dict[str, int]:
    """
    Assign ranks to items in a list.

    Args:
        input_list (List[str]): The list of items.

    Returns:
        Dict[str, int]: A dictionary where keys are items from the input list and
        values are their corresponding ranks (0-based).

    """
    return {item: rank for rank, item in enumerate(input_list)}


def calculate_labelset(
    obs: pd.DataFrame, labelsets: List[str]
) -> Dict[str, Dict[str, Any]]:
    """
    Calculates labelset dictionary based on the provided observations.

    Args:
        obs (pd.DataFrame): DataFrame containing observations.
        labelsets (List[str]): List of labelsets.

    Returns:
        Dict[str, Dict[str, Any]]: A dictionary where keys are labelsets and values are dictionaries containing:
            - "members": set of members for the labelset.
            - "rank": rank of the labelset.
    """
    labelset_dict = {}
    labelset_rank_dict = calculate_labelset_rank(labelsets)
    for item in labelsets:
        labelset_dict.update(
            {
                item: {
                    "members": set(obs[item]),
                    "rank": str(labelset_rank_dict.get(item)),
                }
            }
        )
    return labelset_dict


def add_labelsets_to_cas(cas: Dict[str, Any], labelset_dict: Dict[str, Dict[str, Any]]):
    """
    Updates a CAS dictionary with labelsets derived from a labelset information dictionary.

    Args:
        cas: The CAS dictionary to update.
        labelset_dict: Contains labelset names as keys and dicts with 'rank' (and potentially other info) as values.

    """
    for labelset, value in labelset_dict.items():
        cas.get("labelsets").append(
            {"name": labelset, "description": "", "rank": str(value.get("rank"))}
        )


def get_cell_ids(
    anndata_obs: pd.DataFrame, labelset: str, cell_label: str
) -> List[str]:
    """
    Get cell IDs from an AnnData dataset based on a specified labelset and cell label.

    Args:
        anndata_obs: The observations DataFrame (`obs`) of an AnnData object containing the dataset. This DataFrame
            should include columns for cell type ontology term IDs and cell types.
        labelset: Labelset to filter.
        cell_label: The value of the cell label used to filter rows in `anndata_obs`.

    Returns:
        List[str]: List of cell IDs.
    """
    cell_label_lower = str(cell_label).lower()
    return anndata_obs.index[
        anndata_obs[labelset].astype(str).str.lower() == cell_label_lower
    ].tolist()


def get_cl_annotations_from_anndata(
    anndata_obs: pd.DataFrame, columns_name: str, cell_label: str
) -> Tuple[str, str]:
    """
    Retrieves cell ontology term ID and cell ontology term for a given cell label from the observation DataFrame
        of an AnnData object.

    Args:
        anndata_obs: The observations DataFrame (`obs`) of an AnnData object containing the dataset. This DataFrame
            should include columns for cell type ontology term IDs and cell types.
        columns_name: The name of the column in `anndata_obs` used for filtering based on the cell label.
        cell_label: The value of the cell label used to filter rows in `anndata_obs`.

    Returns:
        tuple: A tuple containing two elements:
            - The first element is the cell type ontology term ID associated with the given cell label.
            - The second element is the cell type (ontology term) associated with the given cell label.
    """
    filtered_df = anndata_obs[anndata_obs[columns_name] == cell_label]

    ontology_term_ids = filtered_df["cell_type_ontology_term_id"].unique().tolist()
    ontology_terms = filtered_df["cell_type"].unique().tolist()

    cell_ontology_term_id = (
        filtered_df["cell_type_ontology_term_id"].iloc[0]
        if len(ontology_term_ids) == 1
        else None
    )
    cell_ontology_term = (
        filtered_df["cell_type"].iloc[0] if len(ontology_terms) == 1 else None
    )

    return cell_ontology_term_id, cell_ontology_term


def generate_parent_cell_lookup(anndata, labelset_dict):
    """
    Generates a lookup dictionary mapping cell labels to various metadata, including cell IDs, rank,
    and cell ontology terms. This function is designed to precompute the lookup information needed for
    CAS annotation generation, especially useful when hierarchy inclusion is desired.

    Args:
        anndata (ad.AnnData): The AnnData object containing the single-cell dataset,
                              including metadata in anndata.obs.
        labelset_dict (Dict[str, Any]): A dictionary where keys are labelset names and values
                                        are dictionaries containing members and their ranks.

    Returns:
        Dict[str, Any]: A dictionary where each key is a cell label and each value is another
                        dictionary containing keys for 'cell_ids' (a set of cell IDs associated
                        with the label), 'rank', 'cell_ontology_term_id', and 'cell_ontology_term'.
    """
    accession_manager = HashAccessionManager()
    parent_cell_look_up = {}
    for k, v in labelset_dict.items():
        for label in v["members"]:
            cell_ontology_term_id, cell_ontology_term = get_cl_annotations_from_anndata(
                anndata.obs, k, label
            )
            cell_ids = get_cell_ids(anndata.obs, k, label)
            cell_set_accession = accession_manager.generate_accession_id(
                cell_ids=cell_ids, labelset=k
            )

            if label in parent_cell_look_up:
                parent_cell_look_up[f"{k}:{label}"]["cell_ids"].update(cell_ids)
            else:
                parent_cell_look_up[f"{k}:{label}"] = {
                    "cell_ids": set(cell_ids),
                    "accession": cell_set_accession,
                    "rank": v.get("rank"),
                    "cell_ontology_term_id": cell_ontology_term_id,
                    "cell_ontology_term": cell_ontology_term,
                }
    return parent_cell_look_up


def update_parent_info(
    value: Dict[str, Any], parent_key: str, parent_value: Dict[str, Any]
):
    """Updates parent information in a child item's dictionary.

    Args:
        value (Dict[str, Any]): The child item's dictionary to be updated.
        parent_key (str): The key of the parent item.
        parent_value (Dict[str, Any]): The parent item's dictionary.

    This function modifies `value` to include `parent` (using `parent_key`),
    `p_accession`, and `parent_rank` based on `parent_value`.
    """
    value.update(
        {
            "parent": parent_key.split(":")[-1],
            "p_accession": parent_value.get("accession"),
            "parent_rank": parent_value.get("rank"),
        }
    )


def add_parent_cell_hierarchy(parent_cell_look_up: Dict[str, Any]):
    """
    Processes parent cell hierarchy information and updates CAS dictionary annotations accordingly.

    Args:
        parent_cell_look_up (Dict[str, Any]): Dictionary containing parent cell information.

    Returns:
        None
    """
    # Establish parent-child relationships
    for (key, value), (inner_key, inner_value) in itertools.product(
        parent_cell_look_up.items(), repeat=2
    ):
        if (
            key == inner_key
            or value.get("parent")
            and value.get("parent_rank", 0) <= inner_value.get("rank")
        ):
            continue

        if value["cell_ids"] != inner_value["cell_ids"] and value["cell_ids"].issubset(
            inner_value["cell_ids"]
        ):
            update_parent_info(value, inner_key, inner_value)
        elif value["cell_ids"] == inner_value["cell_ids"]:
            if int(inner_value["rank"]) < int(value["rank"]):
                update_parent_info(inner_value, key, value)
            elif int(value["rank"]) < int(inner_value["rank"]):
                update_parent_info(value, inner_key, inner_value)
            else:
                raise ValueError(
                    f"{key} and {inner_key} cell labels have the same cell_ids. Cell_ids can't be "
                    f"identical at the same rank."
                )


def add_parent_hierarchy_to_annotations(
    cas: Dict[str, Any], parent_cell_look_up: Dict[str, Any]
):
    """
    Adds parent hierarchy information to annotations in the CAS dictionary.

    Args:
        cas (Dict[str, Any]): The CAS dictionary containing annotations.
        parent_cell_look_up (Dict[str, Any]): Dictionary containing parent cell information.

    Returns:
        None
    """
    annotation_list = cas.get("annotations", [])
    for annotation in annotation_list:
        parent_info = parent_cell_look_up.get(f'{annotation.get("labelset")}:{annotation.get("cell_label")}', {})
        parent = parent_info.get("parent")
        p_accession = parent_info.get("p_accession")
        if parent and p_accession:
            # Add parent data to the annotation
            annotation.update(
                {
                    "parent_cell_set_name": parent,
                    "parent_cell_set_accession": p_accession,
                }
            )
            # Remove redundant CL annotations
            parent_dict = parent_cell_look_up.get(parent, {})
            if parent_dict.get("cell_ontology_term_id") == annotation.get(
                "cell_ontology_term_id"
            ) and parent_dict.get("cell_ontology_term") == annotation.get(
                "cell_ontology_term"
            ):
                annotation.pop("cell_ontology_term_id", None)
                annotation.pop("cell_ontology_term", None)


def json_serializer(obj):
    """
    JSON serializer for objects not serializable by default json code.
    Usage: json.dumps(my_dict, default=json_serializer)

    Args:
        obj: object to serialize

    Returns:
        Serialized object.
    """

    if isinstance(obj, (datetime, date)):
        # return obj.isoformat()
        return obj.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
    raise TypeError("Type %s not serializable" % type(obj))


def get_authors_from_doi(doi):
    """
    Fetches and returns a list of authors from a given DOI (Digital Object Identifier) using the CrossRef API.

    Args:
        doi (str): The DOI of the publication for which to retrieve author information.

    Returns:
        list of dict: A list of dictionaries where each dictionary contains details of one author, including
                      their name ('author_name'), ORCID ID ('orcid'), GitHub username ('github_username'), and email ('email').
                      Each field is a string, and fields without data will be None.

    Raises:
        KeyError: If the author data is not found in the response, indicating a potential issue with the DOI or the data format.

    """
    response = requests.get(f"{CROSSREF_API_URL}{doi}")
    data = response.json()

    try:
        authors = data["message"]["author"]
        author_dict = [
            {
                "author_name": f"{author.get('given')} {author.get('family')}".strip(),
                "orcid": author.get("ORCID"),
                "github_username": author.get("github_username"),
                "email": author.get("email"),
            }
            for author in authors
        ]
        return author_dict
    except KeyError:
        return "Author information not available."



import os

from pathlib import Path
from typing import get_type_hints
from importlib.metadata import version

from cas.file_utils import read_config, read_table_to_dict, write_json_file
from cas.flatten_data_to_tables import serialize_to_tables
from cas.ingest.config_validator import validate
from cas.model import (
    Annotation,
    AnnotationTransfer,
    AutomatedAnnotation,
    CellTypeAnnotation,
    Labelset
)


def ingest_data(
    data_file: str,
    config_file: str,
    out_file: str,
    format: str = "json",
    print_undefined: bool = False,
) -> dict:
    """
    Ingests given data into standard cell annotation schema data structure using the given configuration.

    :param data_file: Unformatted user data in tsv/csv format.
    :param config_file: configuration file path.
    :param out_file: output file path.
    :param format: Data export format. Supported formats are 'json' and 'tsv'
    :param print_undefined: prints null values to the output json if true. Omits undefined values from the json output if
    false. False by default. Only effective in json serialization.
    :return: output data as dict
    """
    cas = ingest_user_data(data_file, config_file)

    if format == "json":
        write_json_file(cas, out_file, print_undefined)
    elif format == "tsv":
        table_name_prefix = os.path.splitext(os.path.basename(data_file))[0]
        if os.path.isfile(out_file):
            out_folder = Path(out_file).parent.absolute()
        else:
            out_folder = out_file
        serialize_to_tables(cas, table_name_prefix, out_folder)

    return cas.to_dict()


def ingest_user_data(data_file: str, config_file: str):
    """
    Ingest given user data into standard cell annotation schema data structure using the given configuration.
    :param data_file: Unformatted user data in tsv/csv format.
    :param config_file: configuration file path.
    """

    config = read_config(config_file)
    is_config_valid = validate(config)
    if not is_config_valid:
        raise Exception("Configuration file is not valid!")
    cas = CellTypeAnnotation(config["author_name"], list(), config["title"])
    cas.description = config.get("description", "")
    cas.cellannotation_schema_version = version("cell-annotation-schema")
    headers, records = read_table_to_dict(data_file, generated_ids=True)
    config_fields = config["fields"]
    labelset_ranks = populate_labelsets(cas, config_fields)
    ao_names = dict()
    utilized_columns = set()
    for record_index in records:
        record = records[record_index]
        if not all(value == '' for value in record.values()):  # skip empty rows
            ao = Annotation("", "")
            parents = [None] * 10
            for field in config_fields:
                # handle hierarchical columns
                if field["column_type"] == "cluster_name":
                    ao.labelset = field["column_name"]
                    ao.cell_label = str(record[field["column_name"]])
                    utilized_columns.add(field["column_name"])
                elif field["column_type"] == "cluster_id":
                    ao.cell_set_accession = str(record[field["column_name"]])
                    ao.rank = int(str(field["rank"]).strip())
                    utilized_columns.add(field["column_name"])
                elif field["column_type"] == "cell_set":
                    if record[field["column_name"]]:
                        parent_ao = get_annotation(ao_names, field, record)
                        register_parent(field, labelset_ranks, parent_ao, parents)
                    utilized_columns.add(field["column_name"])
                else:
                    # handle annotation columns
                    if "typing.List[str]" in str(get_type_hints(ao)[field["column_type"]]):
                        list_value = str(record[field["column_name"]]).split(",")
                        stripped = list(map(str.strip, list_value))
                        setattr(ao, field["column_type"], stripped)
                    else:
                        setattr(ao, field["column_type"], record[field["column_name"]])
                    utilized_columns.add(field["column_name"])

        add_user_annotations(ao, headers, record, utilized_columns)
        add_parent_node_names(ao, ao_names, cas, parents)

        ao_names[ao.cell_label] = ao
        cas.add_annotation_object(ao)
    return cas


def register_parent(field, labelset_ranks, parent_ao, parents):
    """
    Registers the parent annotation object to the parents list.
    Args:
        field: config field
        labelset_ranks: labelset ranks dictionary
        parent_ao: parent to add
        parents: sparse parents list
    """
    if parent_ao.labelset and parent_ao.labelset != field["column_name"]:
        # handle annotations existing in the multiple ranks, only keep the highest rank
        if labelset_ranks[parent_ao.labelset] < int(str(field["rank"]).strip()):
            parents.insert(labelset_ranks[parent_ao.labelset], None)
            parents.insert(int(str(field["rank"]).strip()), parent_ao)
    else:
        parents.insert(int(str(field["rank"]).strip()), parent_ao)


def get_annotation(ao_names, field, record):
    """
    Creates a annotation object if it does not exist in the ao_names dictionary.
    Args:
        ao_names: list of existing annotation objects
        field: config field
        record: data record

    Returns: annotation object
    """
    if record[field["column_name"]] in ao_names:
        ao = ao_names[record[field["column_name"]]]
    else:
        ao = Annotation(field["column_name"], record[field["column_name"]])
    ao.labelset = field["column_name"]
    return ao


def add_user_annotations(ao, headers, record, utilized_columns):
    """
    Adds user annotations that are not supported by the standard schema.
    :param ao: current annotation object
    :param headers: all column names of the user data
    :param record: a record in the user data
    :param utilized_columns: list of processed columns
    """
    not_utilized_columns = [
        column_name for column_name in headers if column_name not in utilized_columns
    ]
    for not_utilized_column in not_utilized_columns:
        if record[not_utilized_column]:
            ao.add_user_annotation(not_utilized_column, record[not_utilized_column])


def add_parent_node_names(ao, ao_names, cas, parents):
    """
    Creates parent nodes if necessary and creates a cluster hierarchy through assigning parent_node_names.
    :param ao: current annotation object
    :param ao_names: list of all created annotation objects
    :param cas: main object
    :param parents: list of current annotation object's parents
    """
    if parents:
        # get first non-null parent
        direct_parent = [x for x in parents if x][0]
        ao.parent_cell_set_name = direct_parent.cell_label
        prev = None
        for parent in reversed(parents):
            if parent:
                if prev:
                    if parent.parent_cell_set_name and parent.parent_cell_set_name != prev.cell_label:
                        print(
                            "Annotation {} has multiple parents: {} and {}".format(parent.cell_label, parent.parent_cell_set_name, prev.cell_label)
                        )
                    if parent.cell_label != prev.cell_label:  # avoid self-references
                        parent.parent_cell_set_name = prev.cell_label
                prev = parent
                if parent.cell_label not in ao_names:
                    cas.add_annotation_object(parent)
                    ao_names[parent.cell_label] = parent


def populate_labelsets(cas, config_fields):
    """
    Populates labelsets list based on the fields of the config.
    :param cas: main object
    :param config_fields: config file fields
    :return: ranks of the labelsets
    """
    labelsets = list()
    ranks = dict()
    for field in config_fields:
        if field["column_type"] == "cell_set" or field["column_type"] == "cluster_name":
            label_set = Labelset(field["column_name"])
            if "rank" in field:
                label_set.rank = str(field["rank"])
                ranks[field["column_name"]] = int(field["rank"])
            labelsets.append(label_set)
    if labelsets:
        cas.labelsets = labelsets
    return ranks



import abc


class BaseAccessionManager(metaclass=abc.ABCMeta):
    """
    Abstract Accession ID generator.
    """

    @abc.abstractmethod
    def generate_accession_id(self, id_recommendation: str = None, labelset: str = None) -> str:
        """
        Generates an auto-increment based accession id. If the recommended accession_id is available, uses it.
        Params:
            id_recommendation: accession id recommendation. Function uses this id if it is available,
            provides an auto-incremented id otherwise.
            labelset: Labelset name. If provided, uses it as a prefix to the accession id.
        Return: accession_id
        """
        pass



import argparse
import os
import pathlib
import warnings

from cas.abc_cas_converter import abc2cas, cas2abc
from cas.add_author_annotations import add_author_annotations_from_file
from cas.anndata_conversion import merge
from cas.anndata_to_cas import anndata2cas
from cas.flatten_data_to_anndata import flatten
from cas.populate_cell_ids import populate_cell_ids
from cas.spreadsheet_to_cas import spreadsheet2cas
from cas.validate import validate as schema_validate
from cas.cas_to_rdf import export_to_rdf
from cas.cas_splitter import split_cas_to_file

warnings.filterwarnings("always")


def main():
    parser = argparse.ArgumentParser(
        prog="cas", description="Cell Type Annotation Tools cli interface."
    )
    subparsers = parser.add_subparsers(help="Available ctat actions", dest="action")

    create_merge_operation_parser(subparsers)
    create_flatten_operation_parser(subparsers)
    create_spreadsheet2cas_operation_parser(subparsers)
    create_anndata2cas_operation_parser(subparsers)
    create_abc2cas_operation_parser(subparsers)
    create_cas2abc_operation_parser(subparsers)
    create_populate_cells_operation_parser(subparsers)
    create_schema_validation_operation_parser(subparsers)
    create_cas2rdf_operation_parser(subparsers)
    create_add_author_annotations_parser(subparsers)
    create_split_cas_parser(subparsers)

    args = parser.parse_args()

    if args.action == "merge":
        json_file_path = args.json
        anndata_file_path = args.anndata
        output_file_path = args.output
        validate = args.validate

        if anndata_file_path == output_file_path:
            raise ValueError("--anndata and --output cannot be the same")

        merge(json_file_path, anndata_file_path, validate, output_file_path)
    elif args.action == "flatten":
        args = parser.parse_args()
        json_file_path = args.json
        anndata_file_path = args.anndata
        output_file_path = args.output

        if output_file_path and os.path.abspath(anndata_file_path) == os.path.abspath(
            output_file_path
        ):
            raise ValueError("--anndata and --output cannot be the same")
        flatten(json_file_path, anndata_file_path, output_file_path)
    elif args.action == "spreadsheet2cas":
        args = parser.parse_args()
        spreadsheet_file_path = args.spreadsheet
        sheet_name = args.sheet
        anndata_file_path = args.anndata
        labelsets = args.labelsets
        schema_name = args.schema
        output_file_path = args.output

        spreadsheet2cas(
            spreadsheet_file_path,
            sheet_name,
            anndata_file_path,
            labelsets,
            schema_name,
            output_file_path,
        )
    elif args.action == "anndata2cas":
        args = parser.parse_args()
        anndata_file_path = args.anndata
        labelsets = args.labelsets
        output_file_path = args.output
        include_hierarchy = args.hierarchy

        anndata2cas(anndata_file_path, labelsets, output_file_path, include_hierarchy)
    elif args.action == "abc2cas":
        args = parser.parse_args()
        cat_set_file_path = args.catset
        cat_file_path = args.cat
        output_file_path = args.output

        abc2cas(cat_set_file_path, cat_file_path, output_file_path)
    # elif args.action == "cas2abc":
    #     args = parser.parse_args()
    #     json_file_path = args.json
    #     cat_set_file_path = args.catset
    #     cat_file_path = args.cat
    #
    #     cas2abc(json_file_path, cat_set_file_path, cat_file_path)
    elif args.action == "populate_cells":
        args = parser.parse_args()
        json_file_path = args.json
        anndata_file_path = args.anndata
        labelsets = None
        if "labelsets" in args and args.labelsets:
            labelsets = [item.strip() for item in str(args.labelsets).split(",")]
        populate_cell_ids(json_file_path, anndata_file_path, labelsets)
    elif args.action == "validate":
        args = parser.parse_args()
        schema = args.schema
        data_file_path = args.data
        schema_validate(schema, data_file_path)
    elif args.action == "cas2rdf":
        args = parser.parse_args()
        export_to_rdf(
            cas_schema=args.schema,
            data=args.data,
            ontology_namespace=args.ontology_ns,
            ontology_iri=args.ontology_iri,
            labelsets=args.labelsets,
            output_path=args.out,
            validate=not args.skip_validate,
            include_cells=not args.exclude_cells,
        )
    elif args.action == "add_author_annotations":
        args = parser.parse_args()
        # Determine the column(s) to use for joining based on the specified arguments
        if args.join_on:
            join_column = args.join_on
        elif args.join_on_cellset_ids:
            join_column = 'cell_set_accession'
        elif args.join_on_labelset_label:
            join_column = ['labelset', 'cell_label']
        else:
            raise ValueError("No valid join column specified.")

        add_author_annotations_from_file(args.cas_json, args.csv, join_column, args.columns, args.output)
    elif args.action == "split_cas":
        cas_json_path = args.cas_json
        split_terms = args.split_on
        multiple_outputs = args.multiple_outputs

        split_cas_to_file(cas_json_path, split_terms, multiple_outputs)


def create_merge_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --json      : Path to the CAS JSON schema file.
    --anndata   : Path to the AnnData file. Ideally, the location will be specified by a resolvable path in the CAS file.
    --validate  : Perform validation checks before writing to the output AnnData file.
    --output    : Output AnnData file name (default: output.h5ad).

    Usage Example:
    --------------
    cd src
    python -m cas merge --json path/to/CAS_schema.json --anndata path/to/input_anndata.h5ad --validate --output path/to/output.h5ad
    """
    parser_merge = subparsers.add_parser(
        "merge",
        description="The CAS and AnnData merge parser",
        help="Test if CAS can be merged to the AnnData and merges if possible.",
    )
    parser_merge.add_argument(
        "--json", required=True, help="Path to the CAS JSON schema file."
    )
    parser_merge.add_argument(
        "--anndata", required=True, help="Path to the AnnData file."
    )
    # TODO find a better argument name and Help message.
    parser_merge.add_argument(
        "-v",
        "--validate",
        action="store_true",
        help="Perform validation checks before writing to the output AnnData file.",
    )
    parser_merge.add_argument(
        "--output",
        default="output.h5ad",
        help="Output AnnData file name (default: output.h5ad).",
    )
    parser_merge.set_defaults(validate=False)


def create_flatten_operation_parser(subparsers):
    """
        Command-line Arguments:
    -----------------------
    --json      : Path to the CAS JSON schema file.
    --anndata   : Path to the AnnData file. Ideally, the location will be specified by a resolvable path in the CAS file.
    --output    : Optional output AnnData file name. If provided a new flatten anndata file will be created,
    otherwise the inputted anndata file will be updated with the flatten data.

    Usage Example:
    --------------
    cd src
    python -m cas flatten --json path/to/json_file.json --anndata path/to/anndata_file.h5ad --output path/to/output_file.h5ad
    """
    parser_flatten = subparsers.add_parser(
        "flatten",
        description="Flattens all content of CAS annotations to an AnnData file.",
        help="Flattens all content of CAS annotations to obs key:value pairs. "
        "Flattens all other content to key_value pairs in uns.",
    )

    parser_flatten.add_argument("--json", required=True, help="Input JSON file path")
    parser_flatten.add_argument(
        "--anndata", required=True, help="Input AnnData file path"
    )
    parser_flatten.add_argument(
        "--output",
        required=False,
        help="Output AnnData file name.",
    )
    parser_flatten.set_defaults(validate=False)


def create_spreadsheet2cas_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --spreadsheet   : Path to the spreadsheet file.
    --sheet         : Target sheet name in the spreadsheet.
    --anndata       : Path to the AnnData file. If not provided, AnnData will be downloaded using CxG LINK in
                    spreadsheet.
    --labelsets     : List of names of observation (obs) fields used to record author cell type names,
        which determine the rank of labelsets in a spreadsheet. If not provided, ranks will be determined based on
        the order of the fields specified in the CELL LABELSET NAME column.
    --schema        : Name of cell annotation schema used to in spreadsheet. It can be one of 'base', 'bican' or 'cap'.
    --output        : Output CAS file name (default: output.json).

    Usage Example:
    --------------
    cd src
    python -m cas spreadsheet2cas --spreadsheet path/to/spreadsheet.xlsx --sheet sheet_name --labelsets item1 item2 item3 --output path/to/output_file.json
    """

    parser_spreadsheet2cas = subparsers.add_parser(
        "spreadsheet2cas",
        description="Converts a spreadsheet to CAS JSON.",
        help="Converts a spreadsheet to Cell Annotation Schema (CAS) JSON.",
        usage="cas spreadsheet2cas --spreadsheet path/to/spreadsheet.xlsx --sheet sheet_name --labelsets item1 item2 item3 --output path/to/output_file.json",
    )

    parser_spreadsheet2cas.add_argument(
        "--spreadsheet", required=True, help="Path to the spreadsheet file."
    )
    parser_spreadsheet2cas.add_argument(
        "--sheet", required=False, help="Target sheet name in the spreadsheet."
    )
    parser_spreadsheet2cas.add_argument(
        "--anndata",
        default=None,
        help="Path to the AnnData file. If not provided, AnnData will be downloaded using CxG LINK in spreadsheet.",
    )
    parser_spreadsheet2cas.add_argument(
        "--labelsets",
        default=None,
        nargs="+",
        help="List of names of observation (obs) fields used to record author cell type names, which determine the "
        "rank of labelsets in a spreadsheet. If not provided, ranks will be determined based on the order of the"
        " fields specified in the CELL LABELSET NAME column.",
    )
    parser_spreadsheet2cas.add_argument(
        "--schema",
        default="cap",
        help="Name of cell annotation schema used to in spreadsheet. It can be one of 'base', 'bican' or 'cap'.",
        choices=["base", "bican", "cap"],
    )
    parser_spreadsheet2cas.add_argument(
        "--output",
        default="output.json",
        help="Output CAS file name (default: output.json).",
    )


def create_anndata2cas_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --anndata       : Path to the AnnData file.
    --labelsets     : List of labelsets, which are names of observation (obs) fields used to record author cell type
    names. The labelsets should be provided in order, starting from rank 0 (leaf nodes) and ascending to higher ranks.

    --output        : Output CAS file name (default: output.json).
    --hierarchy     : Flag indicating whether to include hierarchy in the output.


    Usage Example:
    --------------
    cd src
    python -m cas anndata2cas --anndata path/to/anndata.h5ad --labelsets item1 item2 item3 --output
    path/to/output_file.json
    """
    parser_anndata2cas = subparsers.add_parser(
        "anndata2cas",
        description="Converts an anndata to CAS JSON.",
        help="Converts an anndata to Cell Annotation Schema (CAS) JSON.",
        usage="cas anndata2cas --anndata path/to/anndata.h5ad --labelsets item1 item2 item3 --output path/to/output_file.json",
    )

    parser_anndata2cas.add_argument(
        "--anndata",
        required=True,
        help="Path to the AnnData file.",
    )
    parser_anndata2cas.add_argument(
        "--labelsets",
        required=True,
        nargs="+",
        help="List of labelsets, which are names of observation (obs) fields used to record author cell type names. "
        "The labelsets should be provided in order, starting from rank 0 (leaf nodes) and ascending to higher ranks.",
    )
    parser_anndata2cas.add_argument(
        "--output",
        default="output.json",
        help="Output CAS file name (default: output.json).",
    )
    parser_anndata2cas.add_argument(
        "--hierarchy",
        action="store_true",
        help="Include hierarchy in the output.",
    )


def create_abc2cas_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --catset       : Path to the Cluster Annotation Term Set file.
    --cat           : Path to the Cluster Annotation Term file.
    --output        : Output CAS file name (default: output.json).


    Usage Example:
    --------------
    cd src

    python -m cas abc2cas --catset path/to/cluster_annotation_term_set.csv --cat path/to/cluster_annotation_term.csv
    --output path/to/output_file.json
    """
    parser_abc2cas = subparsers.add_parser(
        "abc2cas",
        description="Converts given ABC cluster_annotation files to CAS JSON.",
        help="Converts given ABC cluster_annotation files to Cell Annotation Schema (CAS) JSON.",
    )

    parser_abc2cas.add_argument(
        "--catset",
        required=True,
        help="Path to the Cluster Annotation Term Set file.",
    )
    parser_abc2cas.add_argument(
        "--cat",
        required=True,
        help="Path to the Cluster Annotation Term file.",
    )
    parser_abc2cas.add_argument(
        "--output",
        default="output.json",
        help="Output CAS file name (default: output.json).",
    )


def create_cas2abc_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --json         : Path to the CAS JSON schema file.
    --catset       : Path to the Cluster Annotation Term Set file.
    --cat          : Path to the Cluster Annotation Term file.


    Usage Example:
    --------------
    cd src

    python -m cas cas2abc --json path/to/json_file.json --catset path/to/cluster_annotation_term_set.csv --cat
    path/to/cluster_annotation_term.csv
    """
    parser_cas2abc = subparsers.add_parser(
        "cas2abc",
        description="Converts given CAS JSON to ABC files. NOTE: This command is under development and may change.",
        help="Converts given Cell Annotation Schema (CAS) to ABC files: cluster_annotation_term and "
        "cluster_annotation_term_set, and writes them to files with cat_file_path and cat_set_file_path. (Under Development)",
    )

    parser_cas2abc.add_argument(
        "--json", required=True, help="Input CAS JSON file path"
    )
    parser_cas2abc.add_argument(
        "--catset",
        required=True,
        help="Path to the Cluster Annotation Term Set file.",
    )
    parser_cas2abc.add_argument(
        "--cat",
        required=True,
        help="Path to the Cluster Annotation Term file.",
    )


def create_populate_cells_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --json      : Path to the CAS JSON schema file.
    --anndata   : Path to the AnnData file. Ideally, the location will be specified by a resolvable path in the CAS file.
    --labelsets : List of labelsets to update with IDs from AnnData. If value is not provided, rank '0' labelset is used.
    The labelsets should be provided in order, starting from rank 0 (leaf nodes) and ascending to higher ranks.

    Usage Example:
    --------------
    cd src
    python -m cas populate_cells --json path/to/json_file.json --anndata path/to/anndata_file.h5ad --labelsets Cluster,Supercluster
    """
    parser_populate = subparsers.add_parser(
        "populate_cells",
        description="The CAS cell IDs population operation.",
        help="Checks for alignment between obs key:value pairs in AnnData file and labelset:cell_label pairs in CAS for some specified list of labelsets. If they are aligned, updates cell_ids in CAS.",
    )

    parser_populate.add_argument("--json", required=True, help="Input JSON file path")
    parser_populate.add_argument(
        "--anndata", required=True, help="Input AnnData file path"
    )
    parser_populate.add_argument(
        "--labelsets",
        help="List of labelsets to update with IDs from AnnData",
        default="",
    )


def create_schema_validation_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --schema    : One of 'base', 'bican' or 'cap'. Identifies the CAS schema to validate data against.
    --data   : Path to the data file (or folder) to validate

    Usage Example:
    --------------
    cd src
    python -m cas validate --schema bican --data path/to/file
    """
    parser_validate = subparsers.add_parser(
        "validate",
        description="The CAS file structure validator",
        help="Test if given CAS is compatible with the CAS schema.",
    )

    parser_validate.add_argument(
        "--schema", required=True, help="Schema name: one of 'base', 'bican' or 'cap'"
    )
    parser_validate.add_argument(
        "--data",
        required=True,
        help="Path to the data file (or folder) to validate",
        type=pathlib.Path,
    )


def create_cas2rdf_operation_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --schema    : (Optional) Name of the CAS release (such as one of `base`, `cap`, `bican`) or path to the
                    CAS schema file or url of the schema file. If not provided, reads the `base` CAS schema from the cas module.
    --data   : Path to the json data file
    --ontology_ns    : Ontology namespace (e.g. `MTG`)
    --ontology_iri    : Ontology IRI (e.g. `https://purl.brain-bican.org/ontology/AIT_MTG/`)
    --labelsets    : (Optional) Labelsets used in the taxonomy (such as `["Cluster", "Subclass", "Class"]`).
    --out    : The output RDF file path.
    --skip_validate    : (Optional) Determines if data-schema validation checks will be performed. Validations are performed by default.
    --exclude_cells    : (Optional) Determines if cell data will be included in the RDF output. Cell data is exported to RDF by default.

    Usage Example:
    --------------
    cd src
    python -m cas cas2rdf --schema bican --data path/to/file.json --ontology_ns MTG --ontology_iri https://purl.brain-bican.org/ontology/AIT_MTG/ --labelsets Cluster Subclass Class --out path/to/output.rdf --exclude_cells
    """
    parser_cas2rdf = subparsers.add_parser(
        "cas2rdf",
        description="CAS to RDF convertor.",
        help="Converts given CAS data into RDF format.",
        usage="cas cas2rdf --schema bican --data path/to/file.json --ontology_ns MTG --ontology_iri https://purl.brain-bican.org/ontology/AIT_MTG/ --labelsets Cluster Subclass Class --out path/to/output.rdf --exclude_cells",

    )

    parser_cas2rdf.add_argument(
        "-s",
        "--schema",
        help="Name of the CAS release (such as one of `base`, `cap`, `bican`) or path to the CAS schema file or url of the schema file. If not provided, reads the `base` CAS schema from the cas module.",
        default="base",
    )
    parser_cas2rdf.add_argument(
        "-d",
        "--data",
        required=True,
        help="Path to the json data file",
        type=pathlib.Path,
    )
    parser_cas2rdf.add_argument(
        "-ns",
        "--ontology_ns",
        required=True,
        help="Ontology namespace (e.g. `MTG`)",
    )
    parser_cas2rdf.add_argument(
        "-iri",
        "--ontology_iri",
        required=True,
        help="Ontology IRI (e.g. `https://purl.brain-bican.org/ontology/AIT_MTG/`)",
    )
    parser_cas2rdf.add_argument(
        "-ls",
        "--labelsets",
        nargs="+",
        help="List of labelsets used in the taxonomy (such as `Cluster Subclass Class`)",
    )
    parser_cas2rdf.add_argument(
        "-o",
        "--out",
        required=True,
        help="The output RDF file path.",
        type=pathlib.Path,
    )
    parser_cas2rdf.add_argument(
        "-sv",
        "--skip_validate",
        action="store_true",
        help="Determines if data-schema validation checks will be performed. Validations are performed by default.",
    )
    parser_cas2rdf.add_argument(
        "-ec",
        "--exclude_cells",
        action="store_true",
        help="Determines if cell data will be included in the RDF output. Cell data is exported to RDF by default.",
    )


def create_add_author_annotations_parser(subparsers):
    """
    Adds a command line parser for the operation to add author annotation fields to CAS JSON files using data from
    a specified CSV file. This operation processes the input CSV and CAS JSON files based on the specified matching
    columns, optionally using selected columns for annotations, and outputs the annotated CAS JSON to a specified file.
    If specific columns are not provided, all columns from the CSV file will be used.

    Command-line Arguments:
    -----------------------
    --cas_json    : Path to the CAS JSON file that will be updated with annotations.
    --csv         : Specifies the path to the CSV file.
    --join_on     : Specifies the single column name in the CSV used for matching records. Each row must have a unique value in this column.
    --join_on_cell_set_id     : Use 'cell_set_id' as the column for matching records.
    --join_on_labelset_label : Use a pair of 'labelset', 'cell_label' columns for matching records.
    --columns     : Optionally specifies which columns in the CSV will be used for annotations. If not provided,
    all columns will be used. Column names containing spaces must be enclosed in quotes (e.g., "Columns Name").
    --output      : Specifies the output file name for the annotated CAS JSON (default: output.json).

    Usage Example:
    --------------
    cd src
    python -m cas add_author_annotations --cas_json path_to_cas.json --csv path_to_csv --join_on CrossArea_cluster --columns random_annotation_x random_annotation_y --output annotated_output.json
    python -m cas add_author_annotations --cas_json path_to_cas.json --csv path_to_csv --join_on_cell_set_id --output annotated_output.json
    python -m cas add_author_annotations --cas_json path_to_cas.json --csv path_to_csv --join_on_labelset_label --output annotated_output.json
    """
    parser_add_annotations = subparsers.add_parser(
        "add_author_annotations",
        description="Add author annotation fields to a CAS JSON file from specified CSV columns. If no columns are specified, all columns are used.",
        help="Adds annotation fields from CSV to a specified CAS JSON file, using all columns if none are specified."
    )
    parser_add_annotations.add_argument(
        "--cas_json", required=True, help="Path to the CAS JSON file to be annotated."
    )
    parser_add_annotations.add_argument(
        "--csv", required=True, help="Path to the CSV file containing annotation data."
    )
    parser_add_annotations.add_argument(
        "--join_on", help="Specifies the single column name in the CSV used for matching records. Each row must have a unique value in this column."
    )
    parser_add_annotations.add_argument(
        "--join_on_cellset_ids", action='store_true', help="Use 'cell_ids' as the column for matching CAS records."
    )
    parser_add_annotations.add_argument(
        "--join_on_labelset_label", action='store_true', help="Use a pair of 'labelset', 'cell_label' columns for matching CAS records."
    )
    parser_add_annotations.add_argument(
        "--columns",
        nargs="+",
        help="Optional space-separated list of column names in the CSV to be added as annotations. All columns are "
             "used if none are specified. Column names containing spaces must be enclosed in quotes (e.g., 'Columns Name')."
    )
    parser_add_annotations.add_argument(
        "--output",
        default="output.json",
        help="Output CAS file name (default: output.json).",
    )


def create_split_cas_parser(subparsers):
    """
    Command-line Arguments:
    -----------------------
    --cas_json          : Path to the CAS JSON file that will be split.
    --split_on          : Cell accession_id(s) to split the CAS file.
    --multiple_outputs  : If set, create multiple output files for each term provided in split_on.
                          If not set, create a single output file containing all child terms.
    Usage Example:
    --------------
    cd src
    python -m cas split_cas --cas_json path/to/cas.json --split_on term1
    python -m cas split_cas --cas_json path/to/cas.json --split_on term1 term2
    python -m cas split_cas --cas_json path/to/cas.json --split_on term1 term2 --multiple_outputs
    """
    parser_split_cas = subparsers.add_parser(
        "split_cas",
        description="Split CAS JSON file based on specified cell label/s.",
        help="Split a CAS JSON file into multiple files based on one or more cell accession_id(s)."
    )
    parser_split_cas.add_argument(
        "--cas_json", required=True, help="Path to the CAS JSON file that will be split"
    )
    parser_split_cas.add_argument(
        "--split_on",
        nargs="+",
        help="Cell accession_id(s) to split the CAS file."
    )
    parser_split_cas.add_argument(
        "--multiple_outputs",
        action="store_true",
        help="If set, create multiple output files for each split_on term; if not set, create a single output file."
    )


if __name__ == "__main__":
    main()



"""
Script Description:
-------------------
This script integrates cell annotations from a CAS (Cell Annotation Schema) JSON file into an AnnData object.
It performs validation checks to ensure data consistency between the CAS file and the AnnData file.
The AnnData file location should ideally be specified as a resolvable path in the CAS file.

Validation Checks:
------------------
1. Check if all barcodes (cell IDs) in the CAS file exist in the AnnData file.
   - If not, a warning is issued with an option to terminate.

2. Check if all barcodes (cell IDs) in the AnnData file exist in the CAS file.
   - If not, a warning is issued with an option to terminate.

3. Check if any obs keys in the AnnData file match labelset names in the CAS.
   - If matches are found:
     - Verify if the cell sets (sets of cell IDs) associated with each annotation for the labelsets in CAS match the AnnData.
     - If they do, check if the cell_labels are identical.
       - If yes, no action is taken.
       - If no, a warning is issued with options to update values to cell_labels in CAS or terminate.
     - If no matches are found:
       - Option to flush and replace obs labelsets with those defined in CAS (labelset:cell_label pairs) or terminate.
         - For example, if a labelset in the AnnData file annotates different sets of cells than those in CAS,
           users have the option to delete the labelset in AnnData and add back the labelset from CAS.

Additional Notes:
-----------------
- JSON data is stored in AnnData.uns (excluding barcodes).
"""

import json
import sys
from typing import Optional

import anndata

from cas.file_utils import read_anndata_file, read_json_file

LABELSET_NAME = "name"

LABELSET = "labelset"

LABELSETS = "labelsets"

ANNOTATIONS = "annotations"

CELL_IDS = "cell_ids"

CELL_LABEL = "cell_label"


def merge(cas_path: str, anndata_path: str, validate: bool, output_file_name: str):
    """
    Tests if CAS json and AnnData are compatible and merges CAS into AnnData if possible.

    Args:
        cas_path: The path to the CAS json file.
        anndata_path: The path to the AnnData file.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
        output_file_name: Output AnnData file name.

    """
    input_json = read_json_file(cas_path)
    input_anndata = read_anndata_file(anndata_path)

    merge_cas_object(input_json, input_anndata, validate, output_file_name)


def merge_cas_object(input_json: dict, input_anndata: Optional[anndata.AnnData], validate: bool, output_file_name: str):
    """
    Tests if CAS json and AnnData are compatible and merges CAS into AnnData if possible.

    Args:
        input_json: The CAS json object.
        input_anndata: The AnnData object.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
        output_file_name: Output AnnData file name.

    """
    test_compatibility(input_anndata, input_json, validate)

    save_cas_to_uns(input_anndata, input_json)
    write_anndata(input_anndata, output_file_name)


def test_compatibility(input_anndata, input_json, validate):
    """
    Tests if CAS and AnnData can be merged.

     Args:
        input_anndata: The AnnData object.
        input_json: The CAS data json object.
        validate: Boolean to determine if validation checks will be performed before writing to the output AnnData file.
    """
    annotations = get_cas_annotations(input_json)
    validate_cell_ids(input_anndata, annotations, validate)

    matching_obs_keys = get_matching_obs_keys(input_anndata, input_json)
    check_labelsets(input_json, input_anndata, matching_obs_keys, validate)


def check_labelsets(cas_json, input_anndata, matching_obs_keys, validate):
    annotations = get_cas_annotations(cas_json)
    derived_cell_ids = get_derived_cell_ids(cas_json)

    for ann in annotations:
        if ann[LABELSET] in matching_obs_keys:
            anndata_labelset_cell_ids = (
                input_anndata.obs.groupby(ann[LABELSET], observed=False)
                .apply(lambda group: set(group.index), include_groups=False)
                .to_dict()
            )
            for cell_label, cell_list in anndata_labelset_cell_ids.items():
                if cell_list == derived_cell_ids.get(
                    str(ann["cell_set_accession"]), set()
                ):
                    handle_matching_labelset(ann, cell_label, input_anndata, validate)
                elif cell_label == ann[CELL_LABEL]:
                    handle_non_matching_labelset(ann, input_anndata, validate, derived_cell_ids)


def get_cas_annotations(input_json):
    return input_json[ANNOTATIONS]


def get_matching_obs_keys(input_anndata, input_json):
    cas_labelset_names = [item[LABELSET_NAME] for item in input_json[LABELSETS]]
    obs_keys = input_anndata.obs_keys()
    matching_obs_keys = list(set(obs_keys).intersection(cas_labelset_names))
    return matching_obs_keys


def handle_matching_labelset(ann, cell_label, input_anndata, validate):
    # Used for label changes
    if cell_label != ann[CELL_LABEL]:
        print(
            f"{ann[CELL_LABEL]} cell ids from CAS match with the cell ids in {cell_label} from anndata. "
            "But they have different cell label."
        )
        if validate:
            sys.exit()
        # add new category to labelset column
        input_anndata.obs[ann[LABELSET]] = input_anndata.obs[
            ann[LABELSET]
        ].cat.add_categories(ann[CELL_LABEL])
        # Overwrite the labelset value with CAS labelset
        input_anndata.obs.loc[ann[CELL_IDS], ann[LABELSET]] = input_anndata.obs.loc[
            ann[CELL_IDS], ann[LABELSET]
        ].map({cell_label: ann[CELL_LABEL]})


def handle_non_matching_labelset(ann, input_anndata, validate, derived_cell_ids):
    # Used for hierarchy changes
    print(
        f"{ann[CELL_LABEL]} cell ids from CAS do not match with the cell ids from anndata. "
        "Please update your CAS json."
    )
    if validate:
        sys.exit()
    # Flush the labelset from anndata
    # input_anndata.obs.loc[list(cell_list), cell_label] = ""
    # Add labelset from CAS to anndata
    cell_ids = derived_cell_ids.get(str(ann["cell_set_accession"]), set())
    input_anndata.obs.loc[list(cell_ids), ann[LABELSET]] = str(ann[CELL_LABEL])


def save_cas_to_uns(input_anndata, input_json):
    # drop cell_ids
    json_without_cell_ids = {
        "author_name": input_json["author_name"],
        "labelset": input_json[LABELSETS],
        "annotations": [
            {key: value for key, value in annotation.items() if key != "cell_ids"}
            for annotation in input_json["annotations"]
        ],
    }
    input_anndata.uns.update({"cas": json.dumps(json_without_cell_ids)})


def validate_cell_ids(input_anndata, annotations, validate):
    # check cell ids
    cas_cell_ids = set()
    for ann in annotations:
        cas_cell_ids.update(ann.get(CELL_IDS, []))
    anndata_cell_ids = set(input_anndata.obs.index)
    # cas -> anndata
    if not cas_cell_ids.issubset(anndata_cell_ids):
        print("Not all members of cell ids from cas exist in anndata.")
        if validate:
            sys.exit()
    # anndata -> cas
    if not anndata_cell_ids.issubset(cas_cell_ids):
        print("Not all members of cell ids from anndata exist in cas.")
        if validate:
            sys.exit()
    return annotations


def write_anndata(input_anndata, output_file_path):
    # Close the AnnData file to prevent blocking
    input_anndata.file.close()
    input_anndata.write(output_file_path)


def get_derived_cell_ids(cas):
    """
    Using the cluster hierarchy derives cell ids for all nodes.
    Args:
        cas: cas json object

    Returns:
        dictionary of cell_set_accession - set of derived cell ids
    """
    derived_cell_ids = dict()

    labelsets = sorted(cas[LABELSETS], key=lambda x: int(x["rank"]))
    for labelset in labelsets:
        ls_annotations = [
            ann for ann in cas[ANNOTATIONS] if ann["labelset"] == labelset["name"]
        ]

        for ann in ls_annotations:
            if "parent_cell_set_accession" in ann:
                cell_ids = set()
                if CELL_IDS in ann and ann[CELL_IDS]:
                    cell_ids = set(ann[CELL_IDS])
                    derived_cell_ids[ann["cell_set_accession"]] = cell_ids
                elif (
                    "cell_set_accession" in ann
                    and ann["cell_set_accession"] in derived_cell_ids
                ):
                    cell_ids = derived_cell_ids[str(ann["cell_set_accession"])]

                if ann["parent_cell_set_accession"] in derived_cell_ids:
                    derived_cell_ids[ann["parent_cell_set_accession"]].update(cell_ids)
                else:
                    derived_cell_ids[ann["parent_cell_set_accession"]] = set(cell_ids)

    return derived_cell_ids



import requests

from typing import Union, Optional, List
from ruamel.yaml import YAML

from cas.file_utils import get_cas_schema, get_cas_schema_names

from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.utils.schema_as_dict import schema_as_dict
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.dumpers import json_dumper

from schema_automator.utils.schemautils import write_schema
from schema_automator.importers.jsonschema_import_engine import JsonSchemaImportEngine

from oaklib.utilities.subsets.value_set_expander import (
    ValueSetExpander,
    ValueSetConfiguration,
)


CAS_ROOT_CLASS = "GeneralCellAnnotationOpenStandard"

CAS_NAMESPACE = "https://cellular-semantics.sanger.ac.uk/ontology/CAS"
CAS_ROOT_NS = "General_Cell_Annotation_Open_Standard"
DEFAULT_PREFIXES = {
    CAS_ROOT_NS: CAS_NAMESPACE + "/",
    "CAS": CAS_NAMESPACE + "/",
    "obo": "http://purl.obolibrary.org/obo/",
    "CL": "http://purl.obolibrary.org/obo/CL_",
    "PCL": "http://purl.obolibrary.org/obo/PCL_",
    "RO": "http://purl.obolibrary.org/obo/RO_",
    "skos": "http://www.w3.org/2004/02/skos/core#",
}


def convert_cas_schema_to_linkml(
    cas_schema: Union[str, dict] = None, output_path: Optional[str] = None
) -> SchemaDefinition:
    """
    Converts a CAS schema to a LinkML schema.
    Args:
        cas_schema (Union[str, dict], optional): Path to the CAS schema file or CAS schema JSON object. If not
        provided, the base CAS schema is read from the CAS module.
        output_path (Optional[str]): The output path for the LinkML schema file. If not specified, no file is written.
    Returns:
        SchemaDefinition: A LinkML SchemaDefinition object.
    """
    if cas_schema is None or cas_schema in get_cas_schema_names().keys():
        cas_schema = get_cas_schema(cas_schema)

    ie = JsonSchemaImportEngine()
    if isinstance(cas_schema, dict):
        schema = ie.loads(
            cas_schema, name="cell-annotation-schema", root_class_name=None
        )
    elif isinstance(cas_schema, str) and (cas_schema.startswith("http://") or cas_schema.startswith("https://")):
        resp = requests.get(cas_schema)
        if resp.status_code == 200:
            schema = ie.loads(
                resp.json(),
                name="cell-annotation-schema",
                root_class_name=None,
            )
        else:
            raise ValueError(f"Failed to fetch schema from the given URL: {cas_schema}"
                             f" with status code: {resp.status_code}")
    else:
        schema = ie.load(
            cas_schema,
            name="cell-annotation-schema",
            format="json",
            root_class_name=None,
        )
    if output_path:
        write_schema(schema, output_path)

    return schema


def decorate_linkml_schema(
    schema_obj: Union[dict, SchemaDefinition],
    ontology_namespace: str,
    ontology_iri: str,
    labelsets: Optional[List[str]] = None,
    output_path: Optional[str] = None,
) -> dict:
    """
    Adds additional properties to the LinkML schema necessary for OWL conversion.
    Args:
        schema_obj (Union[dict, SchemaDefinition]): The LinkML schema object, either as a dictionary or a
        SchemaDefinition.
        ontology_namespace (str): The namespace of the ontology (e.g., 'MTG').
        ontology_iri (str): The ontology IRI (e.g., 'https://purl.brain-bican.org/ontology/AIT_MTG/').
        labelsets (Optional[List[str]]): Labelsets used in the taxonomy, such as ['Cluster', 'Subclass', 'Class'].
        output_path (Optional[str]): Path to the output schema file, if specified.
    Returns:
        dict: The schema decorated with additional properties for OWL conversion.
    """

    if isinstance(schema_obj, SchemaDefinition):
        schema_obj = schema_as_dict(schema_obj)

    schema_obj["id"] = CAS_NAMESPACE

    ontology_namespace = ontology_namespace.upper()
    prefixes = DEFAULT_PREFIXES.copy()
    prefixes["linkml"] = "https://w3id.org/linkml/"
    prefixes["_base"] = ontology_iri
    prefixes[ontology_namespace] = ontology_iri
    labelsets = labelsets or []
    for labelset in labelsets:
        prefixes[labelset] = ontology_iri + f"{labelset}#"

    schema_obj["prefixes"] = prefixes
    schema_obj["default_range"] = "string"
    schema_obj["default_curi_maps"] = ["semweb_context", "obo_context"]
    schema_obj["enums"]["CellTypeEnum"] = {
        "reachable_from": {
            "source_ontology": "obo:cl",
            "source_nodes": ["CL:0000000"],
            "include_self": True,
            "relationship_types": ["rdfs:subClassOf"],
        }
    }
    schema_obj["slots"]["id"] = {"identifier": True, "range": "uriorcurie"}
    schema_obj["slots"]["name"]["slot_uri"] = "rdfs:label"
    schema_obj["slots"]["description"]["slot_uri"] = "IAO:0000115"
    schema_obj["slots"]["cell_label"]["slot_uri"] = "rdfs:label"
    schema_obj["slots"]["cell_fullname"]["slot_uri"] = "skos:preflabel"
    schema_obj["slots"]["cell_ontology_term_id"]["slot_uri"] = "RO:0002473"
    schema_obj["slots"]["cell_ontology_term_id"]["range"] = "CellTypeEnum"
    schema_obj["slots"]["cell_ids"]["slot_uri"] = "CAS:has_cellid"
    schema_obj["slots"]["cell_set_accession"]["identifier"] = True
    schema_obj["slots"]["parent_cell_set_accession"]["slot_uri"] = "RO:0015003"
    schema_obj["slots"]["parent_cell_set_accession"]["range"] = "Annotation"
    schema_obj["slots"]["source_taxonomy"]["range"] = "uriorcurie"
    schema_obj["slots"]["comment"]["slot_uri"] = "IAO:0000115"
    schema_obj["slots"]["labelset"]["slot_uri"] = "CAS:has_labelset"
    schema_obj["slots"]["labelset"]["range"] = "Labelset"
    schema_obj["slots"]["labelsets"]["inlined"] = True
    schema_obj["slots"]["annotations"]["inlined"] = True

    schema_obj["classes"]["Labelset"]["slots"] = list(
        schema_obj["classes"]["Labelset"]["slots"]
    ) + ["id"]
    schema_obj["classes"]["GeneralCellAnnotationOpenStandard"]["slots"] = list(
        schema_obj["classes"]["GeneralCellAnnotationOpenStandard"]["slots"]
    ) + ["id"]
    schema_obj["classes"]["Annotation"]["class_uri"] = "PCL:0010001"

    if output_path:
        write_schema(schema_obj, output_path)

    return schema_obj


def expand_schema(
    config: Optional[str],
    yaml_obj: dict,
    value_set_names: List[str],
    output_path: Optional[str] = None,
):
    """
    Dynamically expands the yaml_obj schema in-place using specified value set names.
    Args:
        config (Optional[str]): Path to the configuration file. If None, a default configuration is used.
        yaml_obj (dict): YAML schema object that will be expanded.
        value_set_names (List[str]): Names of the value sets to be included in the expansion.
        output_path (Optional[str]): Path where the expanded schema file will be saved, if specified.
    Note:
        Source code referenced from: https://github.com/INCATools/ontology-access-kit/blob/main/src/oaklib/utilities/subsets/value_set_expander.py
    """
    expander = ValueSetExpander()
    if config:
        expander.configuration = yaml_loader.load(
            config, target_class=ValueSetConfiguration
        )

    yaml = YAML()
    schema = yaml_loader.load(yaml_obj, target_class=SchemaDefinition)
    if value_set_names is None:
        value_set_names = list(schema.enums.keys())
    for value_set_name in value_set_names:
        if value_set_name not in schema.enums:
            raise ValueError(f"Unknown value set: {value_set_name}")
        value_set = schema.enums[value_set_name]
        pvs = list(expander.expand_value_set(value_set, schema=schema))
        yaml_obj["enums"][value_set_name]["permissible_values"] = {
            str(pv.text): json_dumper.to_dict(pv) for pv in pvs
        }
    if output_path:
        with open(output_path, "w", encoding="UTF-8") as file:
            yaml.dump(yaml_obj, stream=file)
    return yaml_obj



import hashlib
import string
from typing import List

from cas.accession.base_accession_manager import BaseAccessionManager


class HashAccessionManager(BaseAccessionManager):
    def __init__(self, accession_prefix=None, digest_size=5):
        """
        Initializer.
        Params:
            accession_prefix: accession_id prefix
            digest_size: output hash size
        """
        self.accession_prefix = accession_prefix
        self.digest_size = digest_size
        self.accession_ids = list()

    def generate_accession_id(
        self, id_recommendation: str = None, cell_ids: List = None, labelset: str = None
    ) -> str:
        """
        Generates a Blake2b hashing algorithm based hash for the given cell IDs.
        Params:
            id_recommendation: pre-calculated hash accession recommendation. Returns this value if recommendation is a
            valid accession id.
            cell_ids: Cell IDs list. Algorithm sorts cell ids internally.
            labelset: Labelset name. If provided, uses it as a prefix to the accession id.
        Return: accession_id
        """
        if id_recommendation and labelset and ":" not in id_recommendation:
            id_recommendation = labelset + ":" + id_recommendation
        if is_hash_accession(id_recommendation):
            return id_recommendation

        if not cell_ids:
            raise Exception("Cell IDs list is empty.")

        blake_hasher = hashlib.blake2b(
            str.encode(" ".join(sorted(cell_ids))), digest_size=self.digest_size
        )
        accession_id = blake_hasher.hexdigest()
        if labelset:
            accession_id = labelset + ":" + accession_id

        if accession_id in self.accession_ids:
            print("ERROR: Hash ID conflict occurred: " + accession_id)
            # raise Exception("Hash ID conflict occurred: " + accession_id)
        else:
            self.accession_ids.append(accession_id)
        return accession_id


def is_hash_accession(accession_id: str):
    """
    Checks if the given accession is a valid hash accession. Hash accessions are 10 char long and only has hexdigits
    Args:
        accession_id: accession to check

    Returns: True if value is a valid hash accession id, false otherwise.

    """
    if not accession_id:
        return False
    hash_part = accession_id.split(":")[-1]
    return len(hash_part) == 10 and all(c in string.hexdigits for c in hash_part)



from collections import defaultdict
from typing import Any, Dict, List, Union

from cas.file_utils import read_json_file, write_dict_to_json_file

ANNOTATIONS = "annotations"
CELL_LABEL = "cell_label"
CELL_SET_ACCESSION = "cell_set_accession"
LABELSET = "labelset"
LABELSETS = "labelsets"
LABELSETS_NAME = "name"
PARENT_CELL_SET_ACCESSION = "parent_cell_set_accession"


def split_cas_to_file(
    cas_json_path: str, split_terms: Union[List[str], str], multiple_outputs: bool
):
    """
    Splits a CAS JSON file into files based on provided terms, and writes them to disk.

    Args:
        cas_json_path: Path to the CAS JSON file.
        split_terms: Terms used to determine how to split the CAS file; can be a string or a list of strings.
        multiple_outputs: If True, outputs multiple files, one for each split term; otherwise, outputs a single file.

    """
    cas = read_json_file(cas_json_path)
    result = split_cas(cas, split_terms, multiple_outputs)
    if isinstance(result, dict):
        result = [result]
    for idx, cas_item in enumerate(result):
        write_dict_to_json_file(
            (
                f"cas_{split_terms[idx].replace(':', '_')}.json"
                if multiple_outputs
                else "split_cas.json"
            ),
            cas_item,
        )


def split_cas(
    cas: Dict[str, Any], split_terms: Union[List[str], str], multiple_outputs: bool
) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    """
    Splits a CAS dictionary into multiple or single dictionary based on split terms.

    Args:
        cas: Dictionary representing the CAS data.
        split_terms: Terms used to filter and split the CAS data; can be a string or a list of strings.
        multiple_outputs: Determines if the output should be multiple dictionaries or a single dictionary.

    Returns:
        A list of dictionaries if multiple_outputs is True, otherwise a single dictionary.

    Raises:
        ValueError: If any split_terms do not exist in the CAS data under 'parent_cell_set_name'.
    """
    cell_dict = {
        annotation[CELL_SET_ACCESSION]: annotation[PARENT_CELL_SET_ACCESSION]
        for annotation in cas[ANNOTATIONS]
        if (PARENT_CELL_SET_ACCESSION in annotation)
    }
    parent_cell_dict = defaultdict(list)
    for child_cell, parent_cell in cell_dict.items():
        parent_cell_dict[parent_cell].append(child_cell)
    if isinstance(split_terms, str):
        split_terms = [split_terms]
    keys_and_values = list(parent_cell_dict.keys()) + [item for sublist in parent_cell_dict.values() if isinstance(sublist, list) for item in sublist]
    missing_terms = [term for term in split_terms if term not in keys_and_values]
    if missing_terms:
        raise ValueError(
            f"{', '.join(missing_terms)} do not exist in CAS as 'cell_set_name'"
        )
    if multiple_outputs:
        splitted_cas_list = []
        for term in split_terms:
            label_to_copy_list = get_split_terms(parent_cell_dict, term)
            splitted_cas_list.append(
                filter_and_copy_cas_entries(cas, label_to_copy_list)
            )
        return splitted_cas_list
    else:
        label_to_copy_list = get_split_terms(parent_cell_dict, split_terms)
        return filter_and_copy_cas_entries(cas, label_to_copy_list)


def filter_and_copy_cas_entries(
    cas: Dict[str, Any], label_to_copy_list: List[str]
) -> Dict[str, Any]:
    """
    Copies entries from the CAS based on a list of labels to copy.

    Args:
        cas: Dictionary representing the original CAS data.
        label_to_copy_list: List of labels indicating which entries to copy.

    Returns:
        A dictionary with filtered CAS entries.
    """
    output_dict = dict(cas)
    output_dict[ANNOTATIONS] = []
    output_dict[LABELSETS] = []
    labelset_dict = set()
    for annotation in cas[ANNOTATIONS]:
        if annotation[CELL_SET_ACCESSION] in label_to_copy_list:
            output_dict[ANNOTATIONS].append(annotation)
            labelset_dict.add(annotation[LABELSET])
    for labelset in cas[LABELSETS]:
        if labelset[LABELSETS_NAME] in labelset_dict:
            output_dict[LABELSETS].append(labelset)
    return output_dict


def get_split_terms(
    parent_dict: Dict[str, List[str]], split_terms: Union[List[str], str]
) -> List[str]:
    """
    Resolves split terms into a comprehensive list of terms based on a parent-child relationship dictionary.

    Args:
        parent_dict: Dictionary mapping parent terms to lists of child terms.
        split_terms: Initial terms to resolve, can be a string or a list of strings.

    Returns:
        A list of all terms, resolved from the parent_dict.
    """
    if isinstance(split_terms, str):
        split_terms = [split_terms]
    result = []
    stack = list(split_terms)
    while stack:
        term = stack.pop()
        if term in parent_dict:
            stack.extend(parent_dict[term])
        result.append(term)
    return result



from typing import Optional

import anndata

from cas.matrix_file.cxg_resolver import CxGDatasetResolver

CXG_PREFIX = "CellXGene_dataset"


def resolve_matrix_file(
    matrix_file_id: str, cache_folder_path: str = None
) -> Optional[anndata.AnnData]:
    """
    Resolves matrix file identified by the given matrix_file_id.

    Parameters:
        matrix_file_id: dataset identifier
        cache_folder_path: (Optional) matrix file cache folder path
    Returns:
        AnnData object
    """
    protocol = matrix_file_id.split(":")[0]
    dataset_id = matrix_file_id.split(":")[1]

    if protocol == CXG_PREFIX:
        resolver = CxGDatasetResolver(cache_folder_path)
    else:
        raise Exception("Unrecognised matrix file protocol: '{}'".format(protocol))

    return resolver.resolve_matrix_file(dataset_id)


def resolve_matrix_file_path(matrix_file_id: str, cache_folder_path: str = None) -> str:
    """
    Resolves matrix file identified by the given matrix_file_id.

    Parameters:
        matrix_file_id: dataset identifier
        cache_folder_path: (Optional) matrix file cache folder path
    Returns:
        AnnData file path
    """
    protocol = matrix_file_id.split(":")[0]
    dataset_id = matrix_file_id.split(":")[1]

    if protocol == CXG_PREFIX:
        resolver = CxGDatasetResolver(cache_folder_path)
    else:
        raise Exception("Unrecognised matrix file protocol: '{}'".format(protocol))

    return resolver.resolve_matrix_file_path(dataset_id)



import json
import os
import warnings
from importlib import resources

from cas_schema import schema_validator, schemas
from cas.file_utils import get_cas_schema, get_cas_schema_names

warnings.filterwarnings("always")


def validate(schema_name: str, data_path: str):
    """
    Validates all instances in data_path against the given schema.
    Assumes all *.json files in the test_dir should validate against the schema.
    Logs all validation errors and throws an exception if any of the test files is invalid.
    Parameters:
        schema_name: One of 'base', 'bican' or 'cap'. Identifies the CAS schema to validate data against.
        data_path: Path to the data file (or folder) to validate
    """
    schema_name = schema_name.strip().lower()
    schema = get_cas_schema(schema_name)
    if not os.path.exists(data_path):
        raise Exception("Please provide a valid 'data_path': {}".format(data_path))

    result = schema_validator.validate(
        schema, get_cas_schema_names()[schema_name], data_path
    )
    if not result:
        raise Exception("Validation Failed")

    return True



import rdflib
import json

from pathlib import Path
from typing import Union, Optional, List

from cas.linkml_ops.schema import CAS_ROOT_CLASS, DEFAULT_PREFIXES, CAS_NAMESPACE
from cas.file_utils import read_json_file

from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime import SchemaView
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.dumpers import rdflib_dumper
from linkml.validator import Validator
from linkml import generators

CELL_RELATION = "has_cellid"


def dump_to_rdf(
    schema: Union[str, Path, dict],
    instance: Union[str, dict],
    ontology_namespace: str,
    ontology_iri: str,
    labelsets: Optional[List[str]] = None,
    output_path: str = None,
    validate: bool = True,
    include_cells: bool = True,
) -> rdflib.Graph:
    """
    Dumps the given data to an RDF file based on the provided schema.
    Args:
        schema (Union[str, Path, dict]): The schema path, dictionary, or file to be used for RDF generation.
        instance (Union[str, dict]): The data JSON file path or JSON object dictionary.
        ontology_namespace (str): Namespace of the ontology (e.g., `MTG`).
        ontology_iri (str): IRI of the ontology (e.g., `https://purl.brain-bican.org/ontology/AIT_MTG/`).
        labelsets (Optional[List[str]]): Labelsets used in the taxonomy, such as `["Cluster", "Subclass", "Class"]`.
        output_path (Optional[str]):  Path to the output RDF file, if specified.
        validate (bool): Determines if data-schema validation checks will be performed. True by default.
        include_cells (bool): Determines if cell data will be included in the RDF output. True by default.
    Returns:
        An RDFlib graph object.
    """
    schema_def: SchemaDefinition = yaml_loader.load(
        schema, target_class=SchemaDefinition
    )

    if isinstance(instance, str):
        instance = read_json_file(instance)
    instance = remove_empty_strings(instance)

    if validate:
        validate_data(schema_def, instance)
    instance = serialise_author_annotation(instance)

    gen = generators.PythonGenerator(schema_def)
    output = gen.serialize()
    python_module = compile_python(output)
    py_target_class = getattr(python_module, CAS_ROOT_CLASS)

    try:
        py_inst = py_target_class(**instance)
    except Exception as e:
        print(f"Could not instantiate {py_target_class} from the data; exception: {e}")
        return None

    prefixes = DEFAULT_PREFIXES.copy()
    prefixes["_base"] = ontology_iri
    prefixes[ontology_namespace] = ontology_iri
    for labelset in labelsets:
        prefixes[labelset] = ontology_iri + f"{labelset}#"

    g = rdflib_dumper.as_rdf_graph(
        py_inst,
        schemaview=SchemaView(schema_def),
        prefix_map=prefixes,
    )

    add_cl_existential_restrictions(g)
    if not include_cells:
        g.remove((None, rdflib.URIRef(CAS_NAMESPACE + "/" + CELL_RELATION), None))

    if output_path:
        g.serialize(format="xml", destination=output_path)
    return g


def add_cl_existential_restrictions(g: rdflib.Graph):
    """
    Adds existential restrictions to the CL class in the given RDF graph.
    Args:
        g: The RDF graph to be updated.
    """
    sparql_query = """
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX PCL: <http://purl.obolibrary.org/obo/PCL_>
        PREFIX RO: <http://purl.obolibrary.org/obo/RO_>
        
        
        DELETE {
            ?annotation RO:0002473 ?cl_term .
        }
        INSERT { 
            ?annotation rdf:type [
                a  owl:Restriction ;
               owl:onProperty  RO:0002473 ;
               owl:someValuesFrom ?cl_term
            ] .
        }
        WHERE {
          ?annotation a PCL:0010001.
          ?annotation RO:0002473 ?cl_term .
        }
    """
    g.update(sparql_query)


def validate_data(schema: SchemaDefinition, instance: dict) -> bool:
    """
    Validates the given data instance against the specified schema.
    Args:
        schema (SchemaDefinition): The schema to be used for the validation.
        instance (dict): The data instance to be validated.
    Returns:
        bool: Returns `True` if the data is valid; otherwise, logs the validation errors and raises an exception.
    """

    validator = Validator(schema)
    report = validator.validate(instance)
    if report.results:
        print("Validation errors ({}):".format(len(report.results)))
        for result in report.results:
            print(result)
        raise ValueError(
            "Data file is not valid against the schema. {} validation errors found.".format(
                len(report.results)
            )
        )

    print("Data file is valid against the schema.")
    return True


def populate_ids(
    instance: Union[str, Path, dict], ontology_namespace: str, ontology_id: str
) -> dict:
    """
    Populates ID fields in the data instance required for RDF conversion. Updates the instance in-place if it is a
    dictionary.
    Args:
        instance (Union[str, Path, dict]): The data JSON file path or JSON object dictionary.
        ontology_namespace (str): The namespace of the ontology, such as 'MTG'.
        ontology_id (str): The ontology ID to be used for the instance, such as 'AIT_MTG'.
    Returns:
        dict: A JSON object with populated ID properties.
    """

    if isinstance(instance, Path):
        instance = str(instance)
    if isinstance(instance, str):
        data = read_json_file(instance)
        if data is None:
            raise ValueError("No such file: " + instance)
        instance = data

    if "id" in instance and instance["id"]:
        return instance

    if "id" not in instance:
        if "CAS:" not in ontology_id:
            ontology_id = "CAS:" + ontology_id
        instance["id"] = ontology_id

    for labelset in instance.get("labelsets", []):
        if "id" not in labelset:
            labelset["id"] = f"{ontology_namespace}:{labelset['name']}"

    # TODO add id to other properties as well

    return instance


def remove_empty_strings(json_data: dict) -> Union[dict, list]:
    """
    Recursively removes empty strings from the given JSON data.
    Args:
        json_data (dict): The JSON data to be cleaned.
    Returns:
        dict: JSON data with all empty strings removed.
    """
    if isinstance(json_data, dict):
        return {
            key: remove_empty_strings(value)
            for key, value in json_data.items()
            if value != ""
        }
    elif isinstance(json_data, list):
        return [remove_empty_strings(item) for item in json_data if item != ""]
    else:
        return json_data


def serialise_author_annotation(instance: dict) -> dict:
    """
    Author annotations field value is a dict and needs to be serialised to a json string for better representation in RDF.
    Serialises the author annotation data in the instance to a json string.
    Args:
        instance: The instance to be updated.
    Returns:
        The updated instance.
    """
    for annotation in instance.get("annotations", []):
        if "author_annotation_fields" in annotation and annotation["author_annotation_fields"]:
            annotation["author_annotation_fields"] = json.dumps(annotation["author_annotation_fields"])
    return instance



import logging
from typing import Any, Dict, List

import pandas as pd


# Configure logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def validate_markers(
    cas: Dict[str, Any], adata: pd.DataFrame, marker_column: str
) -> bool:
    """
    Validates if the specified marker column in the anndata DataFrame contains all markers mentioned in the
    'annotations' in CAS dictionary. Raises an exception if the marker column does not exist.

    Args:
        cas : A dictionary containing various configurations and annotations, including marker gene evidence.
        adata : An anndata DataFrame.
        marker_column : The name of the column in `adata.var` which contains the list of marker genes to be validated.

    Returns:
        Returns True if all markers are validated without any issue, False or an exception otherwise.

    Raises:
        KeyError: If the specified `marker_column` is not found in `adata.var`.

    Note:
        This function uses the `validate_labelset_markers` to perform the actual validation per annotation entry in
        `cas`.
    """
    if marker_column not in adata.var.keys():
        raise KeyError(f"{marker_column} key does not exist in anndata.var!")

    marker_list = adata.var[marker_column].tolist()
    annotations = cas["annotations"]

    for annotation in annotations:
        validate_labelset_markers(annotation, marker_list)
    return True


def validate_labelset_markers(annotation: Dict[str, Any], marker_list: List[str]):
    """
    Validates if the markers from a specific annotation are present in the provided marker list. Logs a warning if
    any markers are missing.

    Args:
        annotation : A single annotation entry from CAS.
        marker_list : A list of marker genes to be checked against the markers in the annotation.

    Returns:
        This function does not return a value but will log a warning if validation fails.

    Note:
        This function is intended to be used within `validate_markers` to handle individual annotation validation.
    """
    marker_gene_evidence_list = (
        annotation.get("marker_gene_evidence", [])
    )

    if not marker_gene_evidence_list or marker_gene_evidence_list == [""]:
        return None

    marker_gene_evidence_set = set(marker_gene_evidence_list)
    marker_list_set = set(marker_list)

    # Determine the missing marker genes
    missing_markers = marker_gene_evidence_set - marker_list_set

    if missing_markers:
        logger.warning(
            f"Not all marker genes from {annotation['labelset']}-{annotation['cell_label']} pair exist in anndata's var section. "
            f"Missing markers: {', '.join(missing_markers)}"
        )



import abc
from typing import Optional

import anndata


class BaseMatrixFileResolver(metaclass=abc.ABCMeta):
    """
    Base abstract matrix file Resolver
    """

    @abc.abstractmethod
    def resolve_matrix_file(self, dataset_id) -> Optional[anndata.AnnData]:
        """
        Resolves matrix file identified by the given dataset_id.
        Parameters:
            dataset_id: dataset identifier
        Returns:
            AnnData object
        """
        pass

    @abc.abstractmethod
    def resolve_matrix_file_path(self, dataset_id) -> str:
        """
        Resolves matrix file identified by the given dataset_id and returns its path.
        Parameters:
            dataset_id: dataset identifier
        Returns:
            AnnData file path
        """
        pass



import os
from os.path import isfile, join
from typing import Optional

import anndata

from cas.file_utils import read_anndata_file
from cas.matrix_file.base_resolver import BaseMatrixFileResolver

H5AD_FILE_EXTENSION = ".h5ad"


class CxGDatasetResolver(BaseMatrixFileResolver):
    """
    CellxGene dataset resolver.
    """

    def __init__(self, cache_folder_path):
        """
        Initializer.
        Params:
            cache_folder_path: matrix file cache folder path
        """
        self.cache_folder_path = cache_folder_path
        self.cached_datasets = list_cached_datasets(cache_folder_path)

    def resolve_matrix_file(self, dataset_id) -> Optional[anndata.AnnData]:
        return read_anndata_file(self.resolve_matrix_file_path(dataset_id))

    def resolve_matrix_file_path(self, dataset_id) -> str:
        dataset_file_name = dataset_id + H5AD_FILE_EXTENSION
        if dataset_file_name in self.cached_datasets:
            return os.path.join(self.cache_folder_path, dataset_file_name)

        # TODO implement dataset download operation
        raise Exception(
            "Dataset could not be found at location: '{}'".format(
                os.path.join(self.cache_folder_path, dataset_file_name)
            )
        )


def list_cached_datasets(cache_folder_path):
    if os.path.isdir(cache_folder_path):
        return [
            f
            for f in os.listdir(cache_folder_path)
            if os.path.isfile(os.path.join(cache_folder_path, f))
        ]
    else:
        raise Exception(
            "CellxGene dataset cache folder is not accessible: '{}'".format(
                cache_folder_path
            )
        )



import json
from typing import Any, Dict, List

from cas.file_utils import read_anndata_file
from cas.utils.conversion_utils import (
    add_labelsets_to_cas,
    add_parent_cell_hierarchy,
    add_parent_hierarchy_to_annotations,
    calculate_labelset,
    get_authors_from_doi,
    generate_parent_cell_lookup,
)


def anndata2cas(
    anndata_file_path: str,
    labelsets: List[str],
    output_file_path: str,
    include_hierarchy: bool,
):
    """
    Convert an AnnData file to Cell Annotation Schema (CAS) JSON.

    Args:
        anndata_file_path (str): Path to the AnnData file.
        labelsets (List[str]): List of labelsets, which are names of observation (obs) fields used to record author
        cell type names. The labelsets should be provided in order, starting from rank 0 (leaf nodes) and ascending
        to higher ranks.
        output_file_path (str): Output CAS file name.
        include_hierarchy (bool): Flag indicating whether to include hierarchy in the output.
    """

    anndata = read_anndata_file(anndata_file_path)

    labelset_dict = calculate_labelset(anndata.obs, labelsets)

    cas = generate_cas_metadata(dict(anndata.uns))

    add_labelsets_to_cas(cas, labelset_dict)

    parent_cell_look_up = generate_parent_cell_lookup(anndata, labelset_dict)

    add_annotations_to_cas(cas, labelset_dict, parent_cell_look_up)

    if include_hierarchy:
        add_parent_cell_hierarchy(parent_cell_look_up)
        add_parent_hierarchy_to_annotations(cas, parent_cell_look_up)

    # Write the JSON data to the file
    with open(output_file_path, "w") as json_file:
        json.dump(cas, json_file, indent=2)


def generate_cas_metadata(uns: Dict[str, Any]) -> Dict[str, Any]:
    """
    Generates CAS metadata based on the provided 'uns' dictionary.

    Args:
        uns (Dict[str, Any]): The 'uns' dictionary containing metadata.

    Returns:
        Dict[str, Any]: The generated CAS metadata dictionary.
    """
    # TODO None values will be calculated later on
    matrix_file_id = None
    cellannotation_schema_version = uns["schema_version"]
    cellannotation_timestamp = None
    cellannotation_version = None
    cellannotation_url = None
    title = uns.get("title")
    author_name = (
        "John Doe"  # Adding default author_name as it is required in the schema
    )
    author_contact = None
    orcid = None
    cas_init = {
        "matrix_file_id": matrix_file_id,
        "cellannotation_schema_version": cellannotation_schema_version,
        "cellannotation_timestamp": cellannotation_timestamp,
        "cellannotation_version": cellannotation_version,
        "cellannotation_url": cellannotation_url,
        "title": title,
        "author_name": author_name,
        "author_contact": author_contact,
        "orcid": orcid,
        "annotations": [],
        "labelsets": [],
    }
    # Exclude keys with None values
    cas = {k: v for k, v in cas_init.items() if v is not None}
    return cas


def add_annotations_to_cas(
    cas: Dict[str, Any],
    labelset_dict: Dict[str, Any],
    parent_cell_look_up: Dict[str, Any],
):
    """
    Generates CAS annotations based on the provided AnnData object and updates the CAS
    dictionary with new annotations. This function can optionally use a precomputed
    parent cell lookup dictionary to enrich the annotations with hierarchical information.

    Args:
        cas (Dict[str, Any]): The CAS dictionary to be updated with annotations. Expected to have a key
            'annotations' where new annotations will be appended.
        labelset_dict (Dict[str, Any]): A dictionary defining labelsets and their members. This is used to match cell
            labels with their respective metadata and annotations.
        parent_cell_look_up (Dict[str, Any]): A precomputed dictionary containing hierarchical metadata about cell
            labels.

    Returns:
        None: The function directly updates the `cas` dictionary with new annotations. The `parent_cell_look_up` is
        used for enrichment and must be generated beforehand if hierarchical information is to be included.
    """
    for k, v in labelset_dict.items():
        for label in v["members"]:
            labelset = k
            rationale = None
            rationale_dois = None
            marker_gene_evidence = None
            synonyms = None
            category_fullname = None
            category_cell_ontology_exists = None
            category_cell_ontology_term_id = None
            category_cell_ontology_term = None

            anno_init = {
                "labelset": labelset,
                "cell_label": label,
                "cell_fullname": label,
                "cell_set_accession": parent_cell_look_up[f"{labelset}:{label}"]["accession"],
                "cell_ontology_term_id": parent_cell_look_up[f"{labelset}:{label}"][
                    "cell_ontology_term_id"
                ],
                "cell_ontology_term": parent_cell_look_up[f"{labelset}:{label}"]["cell_ontology_term"],
                "cell_ids": list(parent_cell_look_up[f"{labelset}:{label}"]["cell_ids"]),
                "rationale": rationale,
                "rationale_dois": rationale_dois,
                "marker_gene_evidence": marker_gene_evidence,
                "synonyms": synonyms,
                "category_fullname": category_fullname,
                "category_cell_ontology_exists": category_cell_ontology_exists,
                "category_cell_ontology_term_id": category_cell_ontology_term_id,
                "category_cell_ontology_term": category_cell_ontology_term,
            }
            # Exclude keys with None values
            cas.get("annotations").append(
                {k: v for k, v in anno_init.items() if v is not None}
            )



import json
import os
import warnings

from jsonschema import Draft7Validator
from ruamel.yaml import YAML

from cas.file_utils import read_config

SCHEMA_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "./config_schema.yaml"
)


ryaml = YAML(typ="safe")
with open(SCHEMA_PATH) as stream:
    ctat_schema = ryaml.load(stream)
    validator = Draft7Validator(ctat_schema)


def validate(json_object: object) -> bool:
    """
    Validates the given json configuration object using the cell type annotation schema.

    Returns:
    :param json_object: configuration object
    :return: True if object is valid, False otherwise.
    """
    is_valid = True

    if not validator.is_valid(json_object):
        es = validator.iter_errors(json_object)
        for e in es:
            warnings.warn(str(e.message))
            is_valid = False

    return is_valid


def validate_file(file_path: str) -> bool:
    """
    Read the configuration object from the given path and validates it.
    :param file_path: path to the json file
    :return: True if object is valid, False otherwise.
    """
    return validate(read_config(file_path))


def validate_json_str(json_str: str) -> bool:
    """
    Validates the given json string.

    Returns:
    :param json_str: string representation of a json object
    :return: True if object is valid, False otherwise.
    """
    return validate(json.loads(json_str))



from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Any, List, Optional

import dataclasses_json
import pandas as pd
from dataclasses_json import DataClassJsonMixin, config
from dateutil.parser import isoparse
from marshmallow import fields

from cas.reports import get_all_annotations

exclude_none_values = True


class EncoderMixin(DataClassJsonMixin):
    dataclass_json_config = dataclasses_json.config(
        # letter_case=dataclasses_json.LetterCase.CAMEL,
        undefined=dataclasses_json.Undefined.EXCLUDE,
        exclude=lambda f: exclude_none_values and f is None,
    )["dataclasses_json"]


@dataclass
class AutomatedAnnotation(EncoderMixin):
    algorithm_name: str
    """The name of the algorithm used. It MUST be a string of the algorithm's name."""

    algorithm_version: str
    """The version of the algorithm used (if applicable). It MUST be a string of the algorithm's version, which is 
    typically in the format '[MAJOR].[MINOR]', but other versioning systems are permitted (based on the algorithm's 
    versioning)."""

    algorithm_repo_url: str
    """This field denotes the URL of the version control repository associated with the algorithm used (if applicable). 
    It MUST be a string of a valid URL."""

    reference_location: Optional[str]
    """This field denotes a valid URL of the annotated dataset that was the source of annotated reference data. 
    This MUST be a string of a valid URL. The concept of a 'reference' specifically refers to 'annotation transfer' 
    algorithms, whereby a 'reference' dataset is used to transfer cell annotations to the 'query' dataset."""


@dataclass
class Labelset(EncoderMixin):
    name: str
    """name of annotation key"""

    description: Optional[str] = None
    """Some text describing what types of cell annotation this annotation key is used to record"""

    annotation_method: Optional[str] = None
    """The method used for creating the cell annotations. This MUST be one of the following strings: `'algorithmic'`, 
    `'manual'`, or `'both'` """

    automated_annotation: Optional[AutomatedAnnotation] = None
    """A set of fields for recording the details of the automated annotation algorithm used. (Common 'automated 
    annotation methods' would include PopV, Azimuth, CellTypist, scArches, etc.)"""

    rank: Optional[int] = None
    """A number indicating relative granularity with 0 being the most specific.  Use this where a single dataset has 
    multiple keys that are used consistently to record annotations and different levels of granularity."""


@dataclass
class AnnotationTransfer(EncoderMixin):
    transferred_cell_label: Optional[str] = None
    """Transferred cell label"""

    source_taxonomy: Optional[str] = None
    """PURL of source taxonomy."""

    source_node_accession: Optional[str] = None
    """accession of node that label was transferred from"""

    algorithm_name: Optional[str] = None
    """The name of the algorithm used."""

    comment: Optional[str] = None
    """Free text comment on annotation transfer"""


@dataclass
class Review(EncoderMixin):
    """Annotation review."""

    datestamp: Optional[datetime] = field(
        metadata=config(
            encoder=lambda x: (
                x.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z" if x is not None else None
            ),
            decoder=lambda x: isoparse(x) if x is not None else None,
            mm_field=fields.DateTime(format="iso"),
        ),
        default=None,
    )
    """Time and date review was last edited."""

    reviewer: Optional[str] = None
    """Review Author."""

    review: Optional[str] = None
    """Reviewer's verdict on the annotation.  Must be 'Agree' or 'Disagree'."""

    explanation: Optional[str] = None
    """Free-text review of annotation. This is required if the verdict is disagree and should include reasons for disagreement."""


@dataclass
class Annotation(EncoderMixin):
    """
    A collection of fields recording a cell type/class/state annotation on some set os cells, supporting evidence and
    provenance. As this is intended as a general schema, compulsory fields are kept to a minimum. However, tools using
    this schema are encouarged to specify a larger set of compulsory fields for publication.
    Note: This schema deliberately allows for additional fields in order to support ad hoc user fields, new formal
    schema extensions and project/tool specific metadata.
    """

    labelset: str
    """The unique name of the set of cell annotations. 
    Each cell within the AnnData/Seurat file MUST be associated with a 'cell_label' value in order for this to be a 
    valid 'cellannotation_setname'."""

    cell_label: str
    """This denotes any free-text term which the author uses to label cells."""

    cell_set_accession: Optional[str] = None
    """An identifier that can be used to consistently refer to the set of cells being annotated, even if the 
    cell_label changes."""

    cell_fullname: Optional[str] = None
    """This MUST be the full-length name for the biological entity listed in `cell_label` by the author. (If the value 
    in `cell_label` is the full-length term, this field will contain the same value.) \nNOTE: any reserved word used in 
    the field 'cell_label' MUST match the value of this field."""

    cell_ontology_term_id: Optional[str] = None
    """This MUST be a term from either the Cell Ontology or from some ontology that extends it by classifying cell 
    types under terms from the Cell Ontology e.g. the Provisional Cell Ontology."""

    cell_ontology_term: Optional[str] = None
    """This MUST be the human-readable name assigned to the value of 'cell_ontology_term_id"""

    cell_ids: Optional[List[str]] = None
    """List of cell barcode sequences/UUIDs used to uniquely identify the cells"""

    rationale: Optional[str] = None
    """The free-text rationale which users provide as justification/evidence for their cell annotations. 
    Researchers are encouraged to use this field to cite relevant publications in-line using standard academic 
    citations of the form `(Zheng et al., 2020)` This human-readable free-text MUST be encoded as a single string. 
    All references cited SHOULD be listed using DOIs under rationale_dois. There MUST be a 2000-character limit."""

    rationale_dois: Optional[List[str]] = None
    """A list of valid publication DOIs cited by the author to support or provide justification/evidence/context for 
    'cell_label'."""

    marker_gene_evidence: Optional[List[str]] = None
    """List of gene names explicitly used as evidence for this cell annotation."""

    synonyms: Optional[List[str]] = None
    """This field denotes any free-text term of a biological entity which the author associates as synonymous with the 
    biological entity listed in the field 'cell_label'."""

    # TODO modified: added (exclude from json serialisation)
    parent_cell_set_name: Optional[str] = field(default=None, metadata=config(exclude=lambda x: True))

    parent_cell_set_accession: Optional[str] = None
    """A list of accessions of cell sets that subsume this cell set. This can be used to compose hierarchies of 
    annotated cell sets, built from a fixed set of clusters."""

    author_annotation_fields: Optional[dict] = None
    """"A dictionary of author defined key value pairs annotating the cell set. The names and aims of these fields MUST 
    not clash with official annotation fields."""

    transferred_annotations: Optional[List[AnnotationTransfer]] = None

    reviews: Optional[List[Review]] = None

    def add_user_annotation(self, user_annotation_set, user_annotation_label):
        """
        Adds a user defined annotation which is not supported by the standard schema.
        :param user_annotation_set: name of the user annotation set
        :param user_annotation_label: label of the user annotation set
        """
        if not self.author_annotation_fields:
            self.author_annotation_fields = dict()
        self.author_annotation_fields[user_annotation_set] = user_annotation_label


@dataclass
class CellTypeAnnotation(EncoderMixin):

    author_name: str
    """This MUST be a string in the format `[FIRST NAME] [LAST NAME]`"""

    annotations: List[Annotation]
    """A collection of fields recording a cell type/class/state annotation on some set os cells, supporting evidence 
    and provenance. As this is intended as a general schema, compulsory fields are kept to a minimum. However, tools 
    using this schema are encouarged to specify a larger set of compulsory fields for publication."""

    title: str
    """The title of the dataset. This MUST be less than or equal to 200 characters. e.g. 'Human retina cell atlas - 
    retinal ganglion cells'."""

    description: Optional[str] = None
    """The description of the dataset. e.g. 'A total of 15 retinal ganglion cell clusters were identified from over 99K 
    retinal ganglion cell nuclei in the current atlas. Utilizing previous characterized markers from macaque, 5 clusters
     can be annotated.'"""

    matrix_file_id: Optional[str] = None
    """A resolvable ID for a cell by gene matrix file in the form namespace:accession, e.g. 
    CellXGene_dataset:8e10f1c4-8e98-41e5-b65f-8cd89a887122. Please see https://github.com/cellannotation/cell-annotation
    -schema/registry/registry.json for supported namespaces."""

    labelsets: Optional[List[Labelset]] = None
    """A list of labelsets that are used in the annotations."""

    author_contact: Optional[str] = None
    """This MUST be a valid email address of the author"""

    orcid: Optional[str] = None
    """This MUST be a valid ORCID for the author"""

    cellannotation_schema_version: Optional[str] = None
    """The schema version, the cell annotation open standard. Current version MUST follow 0.1.0
    This versioning MUST follow the format `'[MAJOR].[MINOR].[PATCH]'` as defined by Semantic Versioning 2.0.0, 
    https://semver.org/"""

    cellannotation_timestamp: Optional[str] = None
    """The timestamp of all cell annotations published (per dataset). This MUST be a string in the format 
    '%yyyy-%mm-%dd %hh:%mm:%ss'."""

    cellannotation_version: Optional[str] = None
    """The version for all cell annotations published (per dataset). This MUST be a string. The recommended versioning 
    format is `'[MAJOR].[MINOR].[PATCH]'` as defined by Semantic Versioning 2.0.0, https://semver.org/"""

    cellannotation_url: Optional[str] = None
    """A persistent URL of all cell annotations published (per dataset)."""

    author_list: Optional[List[str]] = None
    """This field stores a list of users who are included in the project as collaborators, regardless of their specific 
    role. An example list; John Smith|Cody Miller|Sarah Jones."""

    def add_annotation_object(self, obj):
        """
        Adds given object to annotation objects list
        :param obj: Annotation object to add
        """
        self.annotations.append(obj)

    def set_exclude_none_values(self, value):
        global exclude_none_values
        exclude_none_values = value

    def get_all_annotations(
        self, show_cell_ids: bool = False, labels: list = None
    ) -> pd.DataFrame:
        """
        Lists all annotations.

        Args:
            show_cell_ids: identifies if result have 'cell_ids' column. Default value is false
            labels: list of key(labelset), value(cell_label) pairs to filter annotations
        Returns:
            Annotations data frame
        """
        return get_all_annotations(
            asdict(self), show_cell_ids=show_cell_ids, labels=labels
        )



import pandas as pd


def get_all_annotations(
    cas: dict, show_cell_ids: bool = False, labels: list = None
) -> pd.DataFrame:
    """
    Lists all annotations.

    Args:
        cas: Cell Annotation Schema json object.
        show_cell_ids: identifies if result have 'cell_ids' column. Default value is false
        labels: list of key(labelset), value(cell_label) pairs to filter annotations
    Returns:
        Annotations data frame
    """
    df = pd.json_normalize(cas["annotations"])
    if not show_cell_ids:
        df = df.drop("cell_ids", axis=1, errors="ignore")

    if labels:
        mask = df.apply(lambda row: filter_by_label(row, labels), axis=1)
        return df[mask]
    else:
        return df


def filter_by_label(row, labels):
    """
    Filters and shows row if it
    """
    filter_row = False
    for lbl in labels:
        if row["labelset"] == lbl[0] and row["cell_label"] == lbl[1]:
            filter_row = True
    return filter_row



import csv
import json
import pathlib
from typing import Optional
from importlib import resources

from cas_schema import schemas

import anndata
import numpy as np
from ruamel.yaml import YAML

from cas.model import CellTypeAnnotation


def read_json_file(file_path):
    """
    Reads and parses a JSON file into a Python dictionary.

    Args:
        file_path (str): The path to the JSON file.

    Returns:
        dict: The JSON data as a Python dictionary.

    Returns None if the file does not exist or if there is an issue
    parsing the JSON content.

    Example:
        json_data = read_json_file('path/to/your/file.json')
        if json_data is not None:
            # Use the parsed JSON data as a dictionary
            print(json_data)
    """
    try:
        with open(file_path, "r") as file:
            data = json.load(file)
            return data
    except (FileNotFoundError, json.JSONDecodeError) as e:
        print(f"Error reading JSON file: {e}")
        return None


def read_cas_json_file(file_path: str) -> CellTypeAnnotation:
    """
    Reads and parses a JSON file into a CAS object.

    Args:
        file_path (str): The path to the JSON file.

    Returns:
        dict: The JSON data as a CAS object.
    """
    return CellTypeAnnotation.from_dict(read_json_file(file_path))


def read_cas_from_anndata(anndata_path: str) -> CellTypeAnnotation:
    """
    Reads the CAS json from the anndata uns and parses into a CAS object.
    Args:
        anndata_path: The path to the Anndata file.

    Returns:
        CellTypeAnnotation object.
    """
    input_anndata = read_anndata_file(anndata_path)
    if input_anndata and "cas" in input_anndata.uns:
        return CellTypeAnnotation.from_dict(json.loads(input_anndata.uns["cas"]))
    else:
        raise Exception("Given Anndata file doesn't have a 'cas' object in it's uns.")


def write_json_file(
    cas: CellTypeAnnotation, out_file: str, print_undefined: bool = False
):
    """
    Writes cell type annotation object to a json file.
    :param cas: cell type annotation object to serialize.
    :param out_file: output file path.
    :param print_undefined: prints null values to the output json if true. Omits undefined values from the json output if
    """
    cas.set_exclude_none_values(not print_undefined)

    output_data = cas.to_json(indent=2)
    with open(out_file, "w") as out_file:
        out_file.write(output_data)


def write_dict_to_json_file(output_file_path: str, dictionary: dict):
    with open(output_file_path, "w") as json_file:
        json.dump(dictionary, json_file, indent=2)


def read_anndata_file(file_path: str) -> Optional[anndata.AnnData]:
    """Load anndata object from a file.

    Args:
        file_path: The path to the file containing the anndata object.

    Returns:
        The loaded anndata object if successful, else None.
    """
    try:
        anndata_obj = anndata.read_h5ad(file_path, backed="r")
        return anndata_obj
    except Exception as e:
        print(f"An error occurred while loading the file: {e}")
        return None


def read_table_to_dict(table_path, id_column=0, generated_ids=False):
    """
    Reads table file content into a dict. Key is the first column value and the value is dict representation of the
    Args:
        table_path: Path of the table file
        id_column:  Id column becomes the key of the dict. This column should be unique. Default value is first column.
        generated_ids: If 'True', uses row number as the key of the dict. Initial key is 0.
    Returns:
        Function provides two return values: first; headers of the table and second; the TSV content dict. Key of the
        content is the first column value and the values are dict of row values.
    """
    if table_path.endswith(".tsv"):
        return read_tsv_to_dict(table_path, id_column=id_column, generated_ids=generated_ids)
    elif table_path.endswith(".csv"):
        return read_csv_to_dict(table_path, id_column=id_column, generated_ids=generated_ids)
    else:
        raise Exception("Table file should be either tsv or csv file.")


def read_tsv_to_dict(tsv_path, id_column=0, generated_ids=False):
    """
    Reads tsv file content into a dict. Key is the first column value and the value is dict representation of the
    row values (each header is a key and column value is the value).
    Args:
        tsv_path: Path of the TSV file
        id_column: Id column becomes the key of the dict. This column should be unique. Default value is first column.
        generated_ids: If 'True', uses row number as the key of the dict. Initial key is 0.
    Returns:
        Function provides two return values: first; headers of the table and second; the TSV content dict. Key of the
        content is the first column value and the values are dict of row values.
    """
    return read_csv_to_dict(
        tsv_path, id_column=id_column, delimiter="\t", generated_ids=generated_ids
    )


def read_csv_to_dict(
    csv_path,
    id_column=0,
    id_column_name="",
    delimiter=",",
    id_to_lower=False,
    generated_ids=False,
):
    """
    Reads tsv file content into a dict. Key is the first column value and the value is dict representation of the
    row values (each header is a key and column value is the value).
    Args:
        csv_path: Path of the CSV file
        id_column: Id column becomes the keys of the dict. This column should be unique. Default is the first column.
        id_column_name: Alternative to the numeric id_column, id_column_name specifies id_column by its header string.
        delimiter: Value delimiter. Default is comma.
        id_to_lower: applies string lowercase operation to the key
        generated_ids: If 'True', uses row number as the key of the dict. Initial key is 1.

    Returns:
        Function provides two return values: first; headers of the table and second; the CSV content dict. Key of the
        content is the first column value and the values are dict of row values.
    """
    records = dict()

    headers = []
    with open(csv_path) as fd:
        rd = csv.reader(fd, delimiter=delimiter, quotechar='"')
        row_count = 0
        for row in rd:
            _id = str(row[id_column]).strip()
            if id_to_lower:
                _id = str(_id).strip().lower()

            if generated_ids:
                _id = row_count

            if row_count == 0:
                headers = [str(header).strip() for header in row]
                if id_column_name and id_column_name in headers:
                    id_column = headers.index(id_column_name)
            else:
                row_object = dict()
                for column_num, column_value in enumerate(row):
                    row_object[headers[column_num]] = column_value
                records[_id] = row_object

            row_count += 1

    return headers, records


def read_json_config(file_path: str) -> dict:
    """
    Reads the configuration object from the given path.
    :param file_path: path to the json file
    :return: configuration object (List of data column config items)
    """
    with open(file_path, "r") as fs:
        try:
            return json.load(fs)
        except Exception as e:
            raise Exception("JSON read failed:" + file_path + " " + str(e))


def read_yaml_config(file_path: str) -> dict:
    """
    Reads the configuration object from the given path.
    :param file_path: path to the yaml file
    :return: configuration object (List of data column config items)
    """
    with open(file_path, "r") as fs:
        try:
            ryaml = YAML(typ="safe")
            return ryaml.load(fs)
        except Exception as e:
            raise Exception("Yaml read failed:" + file_path + " " + str(e))


def read_config(file_path: str) -> dict:
    """
    Reads the configuration object from the given path.
    :param file_path: path to the configuration file
    :return: configuration object (List of data column config items)
    """
    file_extension = pathlib.Path(file_path).suffix
    if file_extension == ".json":
        return read_json_config(file_path)
    elif file_extension == ".yaml" or file_extension == ".yml":
        return read_yaml_config(file_path)
    else:
        raise Exception(
            "Given configuration file extension is not supported. "
            "Try a json or yaml file instead of :" + file_path
        )


def update_obs_dataset(obs_dataset, flatten_data):
    """
    Updates obs dataset with flattened data.

    Args:
        obs_dataset (h5py.Dataset): Dataset representing the obs field in the AnnData file.
        flatten_data (dict): Dictionary containing flattened data.
    """
    for key, value in flatten_data.items():
        if key in obs_dataset:
            del obs_dataset[key]
            columns = obs_dataset.attrs["column-order"]
            columns = columns[columns != key]
            obs_dataset.attrs["column-order"] = columns

        obs_dataset.create_dataset(key, data=value.values.astype("O"))

        if key not in obs_dataset.attrs["column-order"]:
            columns = np.append(obs_dataset.attrs["column-order"], key)
            obs_dataset.attrs["column-order"] = columns


def write_json_to_hdf5(group, data):
    """
    Recursively writes JSON-like data to an HDF5 group.

    Args:
        group (h5py.Group): The HDF5 group to write data to.
        data (dict): A dictionary containing the data to be written.

    Returns:
        None
    """
    for key, value in data.items():
        if key in group:
            del group[key]  # Delete the existing key to overwrite it
        if isinstance(value, dict):
            subgroup = group.create_group(key)
            write_json_to_hdf5(subgroup, value)
        elif isinstance(value, list):
            if all(isinstance(item, str) for item in value):
                group.create_dataset(key, data=", ".join(sorted(value)))
            else:
                subgroup = group.create_group(key)
                for i, item in enumerate(value):
                    subgroup.create_dataset(str(i), data=item)
        else:
            group.create_dataset(key, data=value)


def get_cas_schema_names() -> dict:
    """
    Returns the list of available CAS schema names.

    Returns:
        dict: The available CAS schema names.
    """
    return {
        "base": "general_schema.json",
        "cap": "CAP_schema.json",
        "bican": "BICAN_schema.json",
    }


def get_cas_schema(schema_name: Optional[str] = "base") -> dict:
    """
    Reads the schema file from the CAS module and returns as a dictionary.
    Args:
        schema_name: The name of the schema to be returned. Default is 'base'.

    Returns:
        dict: The schema as a dictionary.
    """
    if not schema_name:
        schema_name = "base"

    schema_name = schema_name.strip().lower()
    if schema_name not in get_cas_schema_names():
        raise ValueError(
            "Schema name should be one of: " + ", ".join(get_cas_schema_names().keys())
        )

    schema_file = resources.files(schemas) / get_cas_schema_names()[schema_name]
    with schema_file.open("rt") as f:
        schema = json.loads(f.read())
    return schema



import csv
import os
from dataclasses import asdict

import pandas as pd

from cas.accession.hash_accession_manager import HashAccessionManager, is_hash_accession
from cas.accession.incremental_accession_manager import IncrementalAccessionManager


accession_manager = None


def serialize_to_tables(cta, file_name_prefix, out_folder, project_config):
    """
    Writes cell type annotation object to a series of tsv files.
    Tables to generate:
        - Annotation table (main)
        - Labelset table
        - Metadata
        - Annotation transfer

    Parameters:
        cta: cell type annotation object to serialize.
        file_name_prefix: Name prefix for table names
        out_folder: output folder path.
        project_config: project configuration with extra metadata
    """
    accession_prefix = project_config.get("accession_id_prefix")
    annotation_table_path = generate_annotation_table(accession_prefix, cta, out_folder)
    labelset_table_path = generate_labelset_table(cta, out_folder)
    metadata_table_path = generate_metadata_table(cta, project_config, out_folder)
    annotation_transfer_table_path = generate_annotation_transfer_table(cta, out_folder)
    reviews_table_path = generate_reviews_table(cta, out_folder)
    return [
        annotation_table_path,
        labelset_table_path,
        metadata_table_path,
        annotation_transfer_table_path,
        reviews_table_path,
    ]


def generate_annotation_transfer_table(cta, out_folder):
    """
    Generates annotation transfer table.

    Parameters:
        cta: cell type annotation object to serialize.
        out_folder: output folder path.
    """
    table_path = os.path.join(out_folder, "annotation_transfer.tsv")

    cta = asdict(cta)
    records = list()

    for annotation_object in cta["annotations"]:
        if (
            "cell_set_accession" in annotation_object
            and annotation_object["cell_set_accession"]
            and "transferred_annotations" in annotation_object
            and annotation_object["transferred_annotations"]
        ):
            for ta in annotation_object["transferred_annotations"]:
                record = dict()
                record["target_node_accession"] = annotation_object[
                    "cell_set_accession"
                ]
                record["transferred_cell_label"] = ta.get("transferred_cell_label", "")
                record["source_taxonomy"] = ta.get("source_taxonomy", "")
                record["source_node_accession"] = ta.get("source_node_accession", "")
                record["algorithm_name"] = ta.get("algorithm_name", "")
                record["comment"] = ta.get("comment", "")
                records.append(record)

    if not records:
        record = dict()
        record["target_node_accession"] = ""
        record["transferred_cell_label"] = ""
        record["source_taxonomy"] = ""
        record["source_node_accession"] = ""
        record["algorithm_name"] = ""
        record["comment"] = ""
        records.append(record)

    records_df = pd.DataFrame.from_records(records)
    records_df.to_csv(table_path, sep="\t", index=False)
    return table_path


def generate_metadata_table(cta, project_config, out_folder):
    """
    Generates the metadata table.

    Parameters:
        cta: cell type annotation object to serialize.
        project_config: metadata coming from project config
        out_folder: output folder path.
    """
    table_path = os.path.join(out_folder, "metadata.tsv")

    cta = asdict(cta)
    records = list()

    record = dict()
    record["index"] = "1"
    record["author_name"] = cta.get("author_name", "")
    record["author_contact"] = cta.get("author_contact", "")
    record["orcid"] = project_config.get("author", "")
    record["author_list"] = cta.get("author_list", "")
    record["matrix_file_id"] = project_config.get("matrix_file_id", "")
    record["title"] = cta.get("title", "")
    record["description"] = project_config.get("description", "")
    record["cellannotation_schema_version"] = cta.get(
        "cellannotation_schema_version", ""
    )
    record["cellannotation_timestamp"] = cta.get("cellannotation_timestamp", "")
    record["cellannotation_version"] = cta.get("cellannotation_version", "")
    record["cellannotation_url"] = cta.get("cellannotation_url", "")
    records.append(record)

    records_df = pd.DataFrame.from_records(records)
    records_df.to_csv(table_path, sep="\t", index=False)
    return table_path


def generate_labelset_table(cta, out_folder):
    """
    Generates labelset table.

    Parameters:
        cta: cell type annotation object to serialize.
        out_folder: output folder path.
    """
    table_path = os.path.join(out_folder + "labelset.tsv")

    cta = asdict(cta)
    records = list()

    for labelset in cta["labelsets"]:
        record = dict()
        record["name"] = labelset.get("name", "").replace("_name", "")
        record["description"] = labelset.get("description", "")
        record["rank"] = str(labelset.get("rank", ""))
        record["annotation_method"] = labelset.get("annotation_method", "")
        if "automated_annotation" in labelset and labelset["automated_annotation"]:
            aut_annot = labelset["automated_annotation"]
            name_prefix = "automated_annotation_"
            record[name_prefix + "algorithm_name"] = aut_annot.get("algorithm_name", "")
            record[name_prefix + "algorithm_version"] = aut_annot.get(
                "algorithm_version", ""
            )
            record[name_prefix + "algorithm_repo_url"] = aut_annot.get(
                "algorithm_repo_url", ""
            )
            record[name_prefix + "reference_location"] = aut_annot.get(
                "reference_location", ""
            )
        else:
            name_prefix = "automated_annotation_"
            record[name_prefix + "algorithm_name"] = ""
            record[name_prefix + "algorithm_version"] = ""
            record[name_prefix + "algorithm_repo_url"] = ""
            record[name_prefix + "reference_location"] = ""
        records.append(record)

    records_df = pd.DataFrame.from_records(records)
    records_df.to_csv(table_path, sep="\t", index=False)
    return table_path


def generate_annotation_table(accession_prefix, cta, out_folder):
    """
    Generates annotation table.

    Parameters:
        cta: cell type annotation object to serialize.
        out_folder: output folder path.
        accession_prefix: accession id prefix
    """
    global accession_manager
    std_data_path = os.path.join(out_folder, "annotation.tsv")

    cta = asdict(cta)
    std_records = list()
    std_parent_records = list()
    std_parent_records_dict = dict()

    first_accession = cta["annotations"][0].get("cell_set_accession", "")
    if is_hash_accession(first_accession):
        accession_manager = HashAccessionManager()
    else:
        accession_manager = IncrementalAccessionManager(accession_prefix)
        # sort annotations by accession ids incrementing (if there is)
        if str(first_accession).split(":")[-1].split("_")[-1].isdigit():
            cta["annotations"].sort(
                key=lambda x: (
                    int(str(x["cell_set_accession"]).split(":")[-1].split("_")[-1])
                    if "cell_set_accession" in x
                    and x["cell_set_accession"]
                    and "_" in x["cell_set_accession"]
                    else 0
                )
            )

    id_index = dict()
    for annotation_object in cta["annotations"]:
        record = dict()
        if (
            "cell_set_accession" in annotation_object
            and annotation_object["cell_set_accession"]
        ):
            labelset = str(annotation_object.get("labelset", "")).replace("_name", "")
            record["cell_set_accession"] = accession_manager.generate_accession_id(
                id_recommendation=annotation_object.get("cell_set_accession", ""),
                labelset=labelset,
            )
            annotation_object["cell_set_accession"] = record["cell_set_accession"]
            record["cell_label"] = annotation_object.get("cell_label", "")
            record["cell_fullname"] = annotation_object.get("cell_fullname", "")
            record["parent_cell_set_accession"] = ""
            record["parent_cell_set_name"] = ""
            record["labelset"] = labelset
            record["cell_ontology_term_id"] = annotation_object.get(
                "cell_ontology_term_id", ""
            )
            record["cell_ontology_term"] = annotation_object.get(
                "cell_ontology_term", ""
            )
            record["rationale"] = annotation_object.get("rationale", "")
            record["rationale_dois"] = list_to_string(
                annotation_object.get("rationale_dois", [])
            )
            record["marker_gene_evidence"] = list_to_string(
                annotation_object.get("marker_gene_evidence", [])
            )
            record["synonyms"] = list_to_string(annotation_object.get("synonyms", []))
            if annotation_object.get("author_annotation_fields"):
                for key, value in annotation_object["author_annotation_fields"].items():
                    record[normalize_column_name(key)] = value
            # record["cell_ids"] = annotation_object.get("cell_ids", "")
            std_records.append(record)
            id_index[record["cell_set_accession"]] = record
        else:
            # parent nodes
            parent_label = annotation_object["cell_label"]
            if parent_label not in [
                parent["cell_label"] for parent in std_parent_records
            ]:
                record["cell_set_accession"] = ""
                record["cell_label"] = parent_label
                record["cell_fullname"] = ""
                record["parent_cell_set_accession"] = ""
                record["parent_cell_set_name"] = ""
                record["labelset"] = str(annotation_object.get("labelset", "")).replace(
                    "_name", ""
                )
                std_parent_records.append(record)
        if "parent_cell_set_name" in annotation_object:
            record["parent_cell_set_name"] = annotation_object["parent_cell_set_name"]
            if annotation_object["parent_cell_set_name"] in std_parent_records_dict:
                std_parent_records_dict.get(
                    annotation_object["parent_cell_set_name"]
                ).append(record)
            else:
                children = list()
                children.append(record)
                std_parent_records_dict[annotation_object["parent_cell_set_name"]] = (
                    children
                )
        if "parent_cell_set_accession" in annotation_object:
            record["parent_cell_set_accession"] = annotation_object[
                "parent_cell_set_accession"
            ]
    assign_parent_accession_ids(
        accession_manager, std_parent_records, std_parent_records_dict, cta["labelsets"]
    )
    assign_parent_cell_set_names(id_index)
    std_records.extend(std_parent_records)
    std_records_df = pd.DataFrame.from_records(std_records)
    std_records_df.to_csv(std_data_path, sep="\t", index=False)
    return std_data_path


def generate_reviews_table(cta, out_folder):
    """
    Generates annotation reviews table.

    Parameters:
        cta: cell type annotation object to serialize.
        out_folder: output folder path.
    """
    std_data_path = os.path.join(out_folder, "review.tsv")

    cta = asdict(cta)
    records = list()

    for annotation_object in cta["annotations"]:
        if (
            "cell_set_accession" in annotation_object
            and annotation_object["cell_set_accession"]
            and "reviews" in annotation_object
            and annotation_object["reviews"]
        ):
            labelset = str(annotation_object.get("labelset", "")).replace("_name", "")
            if accession_manager:
                accession = accession_manager.generate_accession_id(
                    id_recommendation=annotation_object.get("cell_set_accession", ""),
                    labelset=labelset,
                )
            else:
                accession = annotation_object.get("cell_set_accession")
            for review in annotation_object["reviews"]:
                record = dict()
                record["target_node_accession"] = accession
                record["datestamp"] = review.get("datestamp", "")
                if record["datestamp"]:
                    # convert time to ISO 8601 format
                    record["datestamp"] = (
                        record["datestamp"].strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
                    )
                record["reviewer"] = review.get("reviewer", "")
                record["review"] = review.get("review", "")
                record["explanation"] = review.get("explanation", "")
                records.append(record)

    if records:
        std_records_df = pd.DataFrame.from_records(records)
        std_records_df.to_csv(std_data_path, sep="\t", index=False)
    else:
        row = ["target_node_accession", "datestamp", "reviewer", "review", "explanation"]
        with open(std_data_path, "w") as f_output:
            tsv_output = csv.writer(f_output, delimiter="\t")
            tsv_output.writerow(row)

    return std_data_path


def list_to_string(my_list: list):
    """
    Converts a list to its string representation. Nanobot has problem with single quotations so removes them as well.
    Parameters:
        my_list: list to serialize
    Returns:
        string representation of the list
    """
    if not my_list:
        str_value = ""
    else:
        str_value = "|".join(my_list)
    return str_value


def assign_parent_accession_ids(
    accession_manager, std_parent_records, std_parent_records_dict, labelsets
):
    """
    Assigns accession ids to parent clusters and updates their references from the child clusters.
    Parameters:
        accession_manager: accession ID generator
        std_parent_records: list of all parents to assign accession ids
        std_parent_records_dict: parent cluster - child clusters dictionary
        labelsets: labelsets list
    """
    label_set_ranks = dict(
        [
            (label_set["name"].replace("_name", ""), label_set["rank"])
            for label_set in labelsets
        ]
    )

    std_parent_records.sort(key=lambda x: int(label_set_ranks[x["labelset"]]))
    for std_parent_record in std_parent_records:
        accession_id = accession_manager.generate_accession_id()
        std_parent_record["cell_set_accession"] = accession_id

        children = std_parent_records_dict.get(std_parent_record["cell_label"], list())
        for child in children:
            child["parent_cell_set_accession"] = accession_id


def assign_parent_cell_set_names(id_index: dict):
    """
    Assigns parent cell set names to the child cell sets.
    Parameters:
        id_index: dictionary of cell set accessions and their corresponding records
    """
    for key, value in id_index.items():
        if value["parent_cell_set_accession"]:
            parent_record = id_index.get(value["parent_cell_set_accession"])
            if parent_record:
                value["parent_cell_set_name"] = parent_record["cell_label"]


def normalize_column_name(column_name: str) -> str:
    """
    Normalizes column name for url compatibility.
    URL compatible column name requirement: All names must match: ^[\w_ ]+$' for to_url()

    Parameters:
        column_name: current column name
    Returns:
        normalized column_name
    """
    return column_name.strip().replace("(", "_").replace(")", "_").replace("-", "_")
